/**
 * Blending operator. Blends shot gathers
 * by padding and adding in order
 * to simulate a simultaneous source
 * survey.
 * @author: Joseph Jennings
 * @version: 2017.02.27
 */

#ifndef BLEND_H_
#define BLEND_H_

#include <vector>
#include "my_operator.h"
#include "hypercube_float.h"
#include "shotpad.h"

class blend : public my_operator {
  public:
    blend(){};
    /**
     * Constructs a blending operator
     * @param daxes the axes of the data.
     *        Expects the first axis to be time.
     *        The second axis to be receiver
     *        The third axis to be shot
     * @param ds the time interval between shots
     * @param stimes array of shot times
     */
    blend(std::vector<axis> daxes, float ds, std::vector<float> stimes);
    blend(std::vector<axis> daxes, float ds, std::vector<float> stimes, bool eshtout);
    /**
     * Constructs a blending operator
     * @param daxes the axes of the data.
     *        Expects the first axis to be time.
     *        The second axis to be receiver
     *        The third axis to be shot
     * @param ds the time interval between shots
     * @param maxdev the maximum amount of deviation from
     *        the regular shot time interval
     * @param mindev the minimum amount of deviation from the
     *        regular time interval
     */
    blend(std::vector<axis> daxes, float ds, float maxdev, float mindev);
    blend(std::vector<axis> daxes, float ds, float maxdev, float mindev, bool eshtout);
    int get_nshots();
    /**
     * Gets the shot times either generated randomly from the constructor
     * or supplied by the user.
     * @return the shot times
     */
    std::vector<float> get_times();
    void get_encshots(hypercube_float *eshts);
    bool forward(bool add, my_vector *mod, my_vector *dat);
    bool adjoint(bool add, my_vector *mod, my_vector *dat);
    virtual ~blend(){
      delete _lsht; delete _osht;
      delete _ubsht; delete _esht; delete _eshts; delete _ubshts;
    };
  private:
    int _nt, _nrx, _nbs, _ns;
    void init(std::vector<axis> daxes, float ds, std::vector<float> stimes, bool eshtout);
    void copy_shot(hypercube_float* sushot, hypercube_float *sishot, int ishot);
    std::vector<float> _stimes;
    std::vector<int> _ubpb;
    float _ds;
    bool _run_nshots = false;
    bool _eshtout;
    hypercube_float *_lsht, *_osht, *_ubsht, *_ubshts, *_esht, *_eshts;
    shotpad _sp;
};

#endif /* BLEND_H_ */
