#include "cgstep.h"
#include <math.h>

/**
 * Allocates memory for a cgstep
 * @param mod the model space
 * @param dat the dataspace
 */
void cgstep::alloc_step(my_vector *mod, norm_vector *dat){
  s  = mod->clone_vec(); s->zero();
  ss = dat->clone_vec(); ss->zero();
  small=1e-20; forget = true;
}

/**
 * Perform one step of conjugate gradient
 * @param m the current model
 * @param g the current gradient
 * @param rr the current residual
 * @param gg the current dataspace gradient (conjugate gradient)
 * @param sc the residual norm after update
 * @return true if sucessfully updated the model
 */
bool cgstep::step_it(my_vector *m, my_vector *g, norm_vector *rr, norm_vector *gg){
  double alfa,beta;
  if(forget){
    /* First iteration: steepest descent */
     beta=0;
     double gg_gg = gg->dot(gg);
     if(fabs(gg_gg)< small) {
        fprintf(stderr,"gg.dot(gg)==%g. Cannot find a proper step size, "
            " will terminate solver.\n",gg_gg);
        return false;
     }
     alfa=-gg->dot(rr)/gg_gg;
  }
  else{
    /* Search a plane (solve a 2X2 matrix) for alfa and beta */
    double gg_gg = gg->dot(gg);
    double ss_ss = ss->dot(ss);
    double gg_ss = gg->dot(ss);
    if(fabs(gg_gg) < small || fabs(ss_ss) < small){
      fprintf(stderr,"gg.dot(gg)==0 or ss.dot(ss)==0. Cannot find a proper step size,"
          " will terminate solver.\n");
     return false;
    }
    double determ = gg_gg * ss_ss * (1.0 - gg_ss/gg_gg * gg_ss/ss_ss);
    double gg_rr = -gg->dot(rr);
    double ss_rr = -ss->dot(rr);
    alfa = ( ss_ss * gg_rr - gg_ss * ss_rr)/determ;
    beta = (-gg_ss * gg_rr + gg_gg * ss_rr)/determ;
  }

  /* Update the solution and residual steps */
  s->scale_add(beta,g,alfa);
  ss->scale_add(beta,gg,alfa);

  /* Update the model and the residual */
  m->add(s);
  rr->add(ss);
 
  forget = false;

  return true;
}
