/*
 * conj_grad.cpp
 *
 *  Created on: Nov 22, 2016
 *      Author: joe
 */

#include "conj_grad.h"
#include "norm_vector.h"
#include "combo_oper.h"
#include "sregf.h"
#include <math.h>
/**
 * Constructs a conjugate gradient solver
 * @param niter number of iterations to run the solver
 */
conj_grad::conj_grad(int niter) {
  _niter = niter;
  if(_niter < 0) seperr("Are you sure want %d iterations?\n",_niter);
}

/**
 * Constructs a conjugate gradient solver
 * @param niter number of iterations to run the solver
 * @param verb flag for turning off verbose output
 */
conj_grad::conj_grad(int niter, bool verb) {
  _niter = niter;
  _verb = verb;
  if(_niter < 0) seperr("Are you sure want %d iterations?\n",_niter);
}

/**
 * Constructs a conjugate gradient solver
 * @param niter number of iterations to run the solver
 * @param fouts string of output tags for writing model,
 *        residual, gradient, etc. files
 */
conj_grad::conj_grad(int niter, std::map<std::string, std::string> fouts) {
  _niter = niter;
  if(_niter < 0) seperr("Are you sure want %d iterations?\n",_niter);
  _fouts = fouts;
  _one = new float[1];
  _step = cgstep();
}

/**
 * Constructs a conjugate gradient solver
 * @param niter
 * @param rtol
 */
conj_grad::conj_grad(int niter, double rtol) {
  _niter = niter; _rtol = rtol;
  if(_niter < 0) seperr("Are you sure want %d iterations?\n",_niter);
  if(_rtol < 0) seperr("Are you sure want %f as the residual tolerance?\n",_rtol);
}

/**
 *
 * @param niter
 * @param rtol
 * @param gtol
 */
conj_grad::conj_grad(int niter, double rtol, double gtol){
  _niter = niter; _rtol = rtol; _gtol = gtol;
  if(_niter < 0) seperr("Are you sure want %d iterations?\n",_niter);
  if(_rtol < 0) seperr("Are you sure want %f as the residual tolerance?\n",_rtol);
  if(_gtol < 0) seperr("Are you sure want %f as the gradient tolerance?\n",_rtol);
}

/**
 *
 * @param gtol
 * @param rtol
 */
conj_grad::conj_grad(double gtol, double rtol) {
  _gtol = gtol; _rtol = rtol;
  if(_gtol < 0) seperr("Are you sure want %f as the gradient tolerance?\n",_gtol);
  if(_rtol < 0) seperr("Are you sure want %f as the residual tolerance?\n",_rtol);
}

/**
 *
 * @param rtol
 */
conj_grad::conj_grad(double rtol) {
  _rtol = rtol;
  if(_rtol < 0) seperr("Are you sure want %f as the residual tolerance?\n",_rtol);
}

/**
 *
 * @param tag
 */
void conj_grad::set_resout(std::string tag) {
  _fouts["rsd"] = tag;
}

/**
 *
 * @param tag
 */
void conj_grad::set_mrsout(std::string tag) {
  _fouts["rsm"] = tag;
}

/**
 *
 * @param tag
 */
void conj_grad::set_grdout(std::string tag) {
  _fouts["grd"] = tag;
}

/**
 *
 * @param tag
 */
void conj_grad::set_grd1out(std::string tag) {
  _fouts["grd1"] = tag;
  if(_fouts["grd"].empty()) {
    _fouts["grd"] = "grddummy";
  }
}

/**
 *
 * @param tag
 */
void conj_grad::set_grd2out(std::string tag) {
  _fouts["grd2"] = tag;
  if(_fouts["grd"].empty()) {
    _fouts["grd"] = "grddummy";
  }
}

/**
 *
 * @param tag
 */
void conj_grad::set_modout(std::string tag) {
  _fouts["mod"] = tag;
}

/**
 *
 * @param tag
 */
void conj_grad::set_mod1out(std::string tag) {
  _fouts["mod1"] = tag;
  if(_fouts["mod"].empty()) {
    _fouts["mod"] = "moddummy";
  }
}

/**
 *
 * @param tag
 */
void conj_grad::set_mod2out(std::string tag) {
  _fouts["mod2"] = tag;
  if(_fouts["mod"].empty()) {
    _fouts["mod"] = "moddummy";
  }
}

/**
 *
 * @param tag
 */
void conj_grad::set_rdnout(std::string tag) {
  if(!_one) _one = new float[1];
  _fouts["rdn"] = tag;
}

/**
 *
 * @param tag
 */
void conj_grad::set_rmnout(std::string tag) {
  if(!_one) _one = new float[1];
  _fouts["rmn"] = tag;
}

/**
 *
 * @param tag
 */
void conj_grad::set_rtnout(std::string tag) {
  if(!_one) _one = new float[1];
  _fouts["rtn"] = tag;
}

/**
 *
 * @param tag
 */
void conj_grad::set_gnout(std::string tag) {
  if(!_one) _one = new float[1];
  _fouts["gdn"] = tag;
}

/**
 *
 * @param tag
 */
void conj_grad::set_ofout(std::string tag) {
  if(!_one) _one = new float[1];
  _fouts["ofn"] = tag;
}


/**
 *
 * @param res
 * @param grd
 * @param mod
 * @param obj
 */
//TODO:
// 1. Generalize the writing of supervectors to files. This needs to be done inside
//    sreg. Just a simple function to write a super vector
// 2. Write out individual gradient norms?
// Also, shorten this function! It is unreadable and poorly designed.
void conj_grad::output(my_vector *res, my_vector *grd, my_vector *mod, double obj) {
  _counter += 1;
  axis a1d = axis(_counter, 0, 1.0, "iteration");
  if(_fouts.size() != 0) {
    /* First check for residuals (data and model)*/
    if(!_fouts["rsd"].empty() || !_fouts["rsm"].empty()) {
      /*  Check if regularized inversion */
      if(!res->name.compare("super_vector")) {
        if(!_fouts["rsd"].empty()) {
          /* Write data space residual */
          super_vector *rv1   = (super_vector*)res;
          hypercube_float *rd = (hypercube_float*)rv1->vecs[0];
          sregf rdout = sregf(rd->get_axes());
          rdout.write_description(_fouts["rsd"]);
          rdout.write_all(_fouts["rsd"],rd);
          rdout.write_axis(_fouts["rsd"],rd->get_ndim()+1,a1d);
        }
        if(!_fouts["rsm"].empty()) {
          /* Write model space residual */
          super_vector *rv2   = (super_vector*)res;
          hypercube_float *rm = (hypercube_float*)rv2->vecs[1];
          sregf rmout = sregf(rm->get_axes());
          rmout.write_description(_fouts["rsm"]);
          rmout.write_all(_fouts["rsm"],rm);
          rmout.write_axis(_fouts["rsm"],rm->get_ndim()+1,a1d);
        }
      } else {
        /* If normal inversion, write data space residual */
        if(!_fouts["rsm"].empty())
          seperr("Cannot write model residual without regularization term.\n");
        hypercube_float *r = (hypercube_float*)res;
        sregf rout = sregf(r->get_axes());
        rout.write_description(_fouts["rsd"]);
        rout.write_all(_fouts["rsd"],r);
        rout.write_axis(_fouts["rsd"],r->get_ndim()+1,a1d);
      }
    }
    /* Check and write gradient */
    if(!_fouts["grd"].empty()) {
      if(!grd->name.compare("super_vector")) {
        if(!_fouts["grd1"].empty()) {
          /* Write gradient for model 1 */
          super_vector *gv1   = (super_vector*)grd;
          hypercube_float *g1 = (hypercube_float*)gv1->vecs[0];
          sregf g1out = sregf(g1->get_axes());
          g1out.write_description(_fouts["grd1"]);
          g1out.write_all(_fouts["grd1"],g1);
          g1out.write_axis(_fouts["grd1"],g1->get_ndim()+1,a1d);
        }
        if(!_fouts["grd2"].empty()) {
          /* Write gradient for model 2 */
          super_vector *gv2   = (super_vector*)grd;
          hypercube_float *g2 = (hypercube_float*)gv2->vecs[1];
          sregf g2out = sregf(g2->get_axes());
          g2out.write_description(_fouts["grd2"]);
          g2out.write_all(_fouts["grd2"],g2);
          g2out.write_axis(_fouts["grd2"],g2->get_ndim()+1,a1d);
        }
      } else {
        hypercube_float *g = (hypercube_float*)grd;
        sregf gout = sregf(g->get_axes());
        gout.write_description(_fouts["grd"]);
        gout.write_all(_fouts["grd"],g);
        gout.write_axis(_fouts["grd"],g->get_ndim()+1,a1d);
      }
    }
    /* Check and write model update */
    if(!_fouts["mod"].empty()) {
      if(!mod->name.compare("super_vector")) {
        if(!_fouts["mod1"].empty()) {
          /* Write updated model 1 */
          super_vector *mv1   = (super_vector*)mod;
          hypercube_float *m1 = (hypercube_float*)mv1->vecs[0];
          sregf m1out = sregf(m1->get_axes());
          m1out.write_description(_fouts["mod1"]);
          m1out.write_all(_fouts["mod1"],m1);
          m1out.write_axis(_fouts["mod1"],m1->get_ndim()+1,a1d);
        }
        if(!_fouts["mod2"].empty()) {
          /* Write updated model 2 */
          super_vector *mv2   = (super_vector*)mod;
          hypercube_float *m2 = (hypercube_float*)mv2->vecs[1];
          sregf m2out = sregf(m2->get_axes());
          m2out.write_description(_fouts["mod2"]);
          m2out.write_all(_fouts["mod2"],m2);
          m2out.write_axis(_fouts["mod2"],m2->get_ndim()+1,a1d);
        }
      } else {
        hypercube_float *m = (hypercube_float*)mod;
        sregf mout = sregf(m->get_axes());
        mout.write_description(_fouts["mod"]);
        mout.write_all(_fouts["mod"],m);
        mout.write_axis(_fouts["mod"],m->get_ndim()+1,a1d);
      }
    }
    /* Check and write residual norm */
    if(!_fouts["rdn"].empty() || !_fouts["rmn"].empty()) {
      if(!res->name.compare("super_vector")) {
        super_vector *rsup = (super_vector*)res;
        if(!_fouts["rdn"].empty()) {
          sregf rnout = sregf();
          std::vector<axis> raxis;
          raxis.push_back(a1d);
          rnout.set_axes(raxis);
          rnout.write_description(_fouts["rdn"]);
          _one[0] = sqrt(rsup->vecs[0]->dot(rsup->vecs[0]));
          if(sizeof(float) != srite(_fouts["rdn"].c_str(),_one, sizeof(float)))
            seperr("Could not write data residual norm\n");
        }
        if(!_fouts["rmn"].empty()) {
          sregf rnout = sregf();
          std::vector<axis> raxis;
          raxis.push_back(a1d);
          rnout.set_axes(raxis);
          rnout.write_description(_fouts["rmn"]);
          _one[0] = sqrt(rsup->vecs[1]->dot(rsup->vecs[1]));
          if(sizeof(float) != srite(_fouts["rmn"].c_str(),_one, sizeof(float)))
            seperr("Could not write model space residual norm\n");
        }
      }
      else {
        if(!_fouts["rmn"].empty())
          seperr("Cannot write model residual without regularization term.\n");
        sregf rnout = sregf();
        std::vector<axis> raxis;
        raxis.push_back(a1d);
        rnout.set_axes(raxis);
        rnout.write_description(_fouts["rdn"]);
        _one[0] = sqrt(res->dot(res));
        if(sizeof(float) != srite(_fouts["rdn"].c_str(),_one, sizeof(float)))
          seperr("Could not write residual norm\n");
      }
    }
    if(!_fouts["rtn"].empty()) {
      sregf rnout = sregf();
      std::vector<axis> raxis;
      raxis.push_back(a1d);
      rnout.set_axes(raxis);
      rnout.write_description(_fouts["rtn"]);
      _one[0] = sqrt(res->dot(res));
      if(sizeof(float) != srite(_fouts["rtn"].c_str(),_one, sizeof(float)))
        seperr("Could not write total residual norm\n");
    }
    /* Check and write gradient norm */
    if(!_fouts["gdn"].empty()) {
      sregf gnout = sregf();
      std::vector<axis> gaxis;
      gaxis.push_back(a1d);
      gnout.set_axes(gaxis);
      gnout.write_description(_fouts["gdn"]);
      _one[0] = sqrt(grd->dot(grd));
      if(sizeof(float) != srite(_fouts["gdn"].c_str(),_one,sizeof(float)))
        seperr("Could not write gradient norm\n");
    }
  }
}

/**
 *
 * @param dat
 * @param mod
 */
void conj_grad::alloc_solve(my_vector *dat, my_vector *mod) {
  _res = (norm_vector*)dat->clone_vec(); _res->zero();
  _drs = (norm_vector*)dat->clone_vec(); _drs->zero();
  _grd = mod->clone_vec(); _grd->zero();
  _step = cgstep(); _step.alloc_step(mod,(norm_vector*)dat);
  _alloced = true;
}
/**
 *
 * @param dat
 * @param mod
 */
void conj_grad::alloc_solvetp(my_vector *dat, my_vector *mod) {
  _res = (norm_vector*)dat->clone_vec(); _res->zero();
  _drs = (norm_vector*)dat->clone_vec(); _drs->zero();
  _grd   = mod->clone_vec(); _grd->zero();
  _grdtp = mod->clone_vec(); _grdtp->zero();
  _step = cgstep(); _step.alloc_step(mod,(norm_vector*)dat);
  _alloced = true;
}

void conj_grad::dealloc_solve() {
  if(_alloced) {
    delete _res; delete _drs; delete _grd;
  }
}

void conj_grad::dealloc_solvetp() {
  if(_alloced) {
    delete _res; delete _drs; delete _grd; delete _grdtp;
  }
}

/**
 * Run the conjugate gradient inversion
 * @param op input operator
 * @param dat input data
 * @param mod inverted model
 */
void conj_grad::run(my_operator *op, my_vector *dat, my_vector *mod) {
  alloc_solve(dat, mod);
  for(int iter = 0; iter < _niter; ++iter) {
    /* First compute the objective function */
    op->forward(false, mod, _res);
    _res->scale_add(1.0,dat,-1.0);
    double f0 = (0.5)*(_res->dot(_res));
    /* Compute the gradient*/
    op->adjoint(false, _grd, _res);
    op->forward(false, _grd, _drs);
    /* Compute the step length and update*/
    if(!_step.step_it(mod,_grd,_res,_drs)) break;
    /* Reevaluate the objective function and output */
    double f1 = (0.5)*_res->dot(_res);
    if(f1 >= f0) {
      fprintf(stderr, "Objective function did not reduce, will terminate solver.\n");
      break;
    }
    output(_res,_grd,mod,f0);
    if(_verb) fprintf(stderr,"iter=\t%d\t obj=\t%g\n",iter+1,f1);
  }
  dealloc_solve();
}

void conj_grad::run(my_operator *op, my_operator *rop, my_vector* dat,
    my_vector* rdat, my_vector* mod, float eps) {
  /* Construct column op and super vec */
  rop->set_scale(eps); rdat->scale(eps);
  col_op *col      = new col_op(op,rop);
  super_vector *sv = new super_vector((norm_vector*)dat,(norm_vector*)rdat);
  /* Run solver */
  run(col,sv,mod);
}

/**
 * Run the conjugate gradient inversion
 * @param op input operator
 * @param grdtap gradient taper operator
 * @param dat input data
 * @param mod inverted model
 */
void conj_grad::run(my_operator *op, my_operator *grdtap, my_vector *dat, my_vector *mod) {
  alloc_solvetp(dat, mod);
  for(int iter = 0; iter < _niter; ++iter) {
    /* First compute the objective function */
    op->forward(false, mod, _res);
    _res->scale_add(1.0,dat,-1.0);
    double f0 = (0.5)*(_res->dot(_res));
    /* Compute the gradient*/
    op->adjoint(false, _grd, _res);
    /* Taper the gradient */
    grdtap->forward(false,_grd,_grdtp);
    /* Compute data space gradient */
    op->forward(false, _grdtp, _drs);
    /* Compute the step length and update*/
    if(!_step.step_it(mod,_grdtp,_res,_drs)) break;
    /* Reevaluate the objective function and output */
    double f1 = (0.5)*_res->dot(_res);
    if(f1 >= f0) {
      fprintf(stderr, "Objective function did not reduce, will terminate solver.\n");
      break;
    }
    output(_res,_grdtp,mod,f0);
    if(_verb) fprintf(stderr,"iter=\t%d\t obj=\t%g\n",iter+1,f1);
  }
  dealloc_solvetp();
}
