/*
 * lininterp.cpp
 *
 *  Created on: Nov 4, 2016
 *      Author: joe
 */

#include "lininterp.h"
#include "hypercube_float.h"
#include <math.h>

/**
 * Constructs a 1D linear interpolation operator
 * @param iaxis the number of the axis along which to interpolate
 */
//FIXME: This does not for any axis work as the
//       order of looping would need to change
lininterp::lininterp(int iaxis) {
  _iax = iaxis;
}

/**
 * Constructs a 1D linear interpolation operator
 * @param coords the input coordinates to be interpolated
 */
lininterp::lininterp(int iaxis, std::vector<float> coords) {
  _iax = iaxis;
  _coords = coords;
}

/**
 * Applies a forward linear interpolation operator
 * @param add flag for adding to the output
 * @param mod the coarsely sampled signal
 * @param dat the more finely sampling signal
 * @return status flag to check the operation
 */
bool lininterp::forward(bool add, my_vector *mod, my_vector *dat) {
  if(!add) dat->zero();
  hypercube_float *m = (hypercube_float*) mod;
  hypercube_float *d = (hypercube_float*) dat;

  float f;
  for(int k = 0; k < d->get_axis(_iax).n; ++k) {
    if(!_coords.empty()) {
      /* Interpolate from coordinates */
      f = (_coords[k] - m->get_axis(_iax).o) / m->get_axis(_iax).d;
    } else {
      /* Interpolate from regularly sampled axis */
      float dcoord = d->get_axis(_iax).o + d->get_axis(_iax).d * k;
      f = (dcoord - m->get_axis(_iax).o) / m->get_axis(_iax).d;
    }
    /* Compute and apply weights */
    int im = floor(f);
    std::pair<float,float> wts = wgts(f);
    wts.first *= scale; wts.second *= scale;
    if(im >= 0 && im < m->get_axis(_iax).n - 1) {
      d->vals[k] += wts.second * m->vals[im] + wts.first * m->vals[im+1];
    }
  }
  return true;
}

/**
 * Applies the adjoint of the linear interpolation operator
 * @param add Flag
 * @param mod
 * @param dat
 * @return a status flag for checking the operation
 */
bool lininterp::adjoint(bool add, my_vector *mod, my_vector *dat) {
  if(!add) mod->zero();
  hypercube_float *m = (hypercube_float*) mod;
  hypercube_float *d = (hypercube_float*) dat;

  float f;
  for(int k = 0; k < d->get_axis(_iax).n; ++k) {
    if(!_coords.empty()) {
      /* Interpolate from coordinates */
      f = (_coords[k] - m->get_axis(_iax).o) / m->get_axis(_iax).d;
    } else {
      /* Interpolate from regularly sampled axis */
      float dcoord = d->get_axis(_iax).o + d->get_axis(_iax).d * k;
      f = (dcoord - m->get_axis(_iax).o) / m->get_axis(_iax).d;
    }
    /* Compute and apply weights */
    int im = floor(f);
    std::pair<float,float> wts = wgts(f);
    wts.first *= scale; wts.second *= scale;
    if(im >= 0 && im < m->get_axis(_iax).n - 1) {
      m->vals[im  ] +=  wts.second * d->vals[k];
      m->vals[im+1] +=  wts.first  * d->vals[k];
      fprintf(stderr,"im=%d d=%f fx=%f gx=%f m=%f mp1=%f\n",im,d->vals[k],wts.first,wts.second,m->vals[im],m->vals[im+1]);
    }
  }
  return true;
}

/**
 * Computes the interpolation weights
 * @param m the coarsely sampled signal
 * @param d the finely sampled signal
 * @param k the index on the coarsely sampled axis
 * @return a tuple containing the model index and the interpolation weights
 */
std::tuple<int,float,float> lininterp::wgts(hypercube_float *m, hypercube_float *d, int k) {
  float dcoord = d->get_axis(_iax).o + d->get_axis(_iax).d * k;
  float f = (dcoord - m->get_axis(_iax).o)/m->get_axis(_iax).d;
  int im = floor(f);
  float fx = f - im; float gx = 1 - fx;
  return std::make_tuple(im, fx, gx);
}

std::pair<float,float> lininterp::wgts(float f) {
  std::pair<float,float> weights;
  int i = floor(f);
  weights.first = f - i;
  weights.second = 1 - weights.first;
  return weights;
}
