/*
 * lv_conv1df.cpp
 *
 *  Created on: Mar 15, 2017
 *      Author: joe
 */

#include "lint_fast.h"
#include <math.h>

lint_fast::lint_fast(int nd, int j, int nlag) {

  /* Compute the number of filters on each axis */
  _nd = nd;
  _nf = static_cast<int>(ceilf(static_cast<float>(_nd-2.0)/static_cast<float>(j)) + 1);
  _nlag = nlag;

  /* Assume lags increase linearly */
  for(int i = 0; i < _nlag; ++i) {
    _lag.push_back(i);
  }

  /* Temporary arrays */
  std::vector<axis> taxes;
  taxes.push_back(axis(_nlag,0.0,1.0));
  _b0 = new hypercube_float(taxes,true);
  _d0 = new hypercube_float(taxes,true);

  /* Construct the array of blocks */
  std::vector<int> b, e;
  find_optimal_sizes(_nd,j,b,e);
  for(int i = 0; i < (int)b.size(); ++i) {
    _blocks.push_back(block1d(b[i],e[i]));
  }

}

lint_fast::lint_fast(int nd, int j, std::vector<int> lag) {
  /* Compute the number of filters on each axis */
  _nd = nd;
  _nf = static_cast<int>(ceilf(static_cast<float>(_nd-2.0)/static_cast<float>(j)) + 1);
  _nlag = lag.size();

  /* Save lags */
  if(!lag.empty()) {
    _lag = lag;
  } else {
    seperr("lv_conv1df: Input lag array is empty. Please initialize with a "
        "non-empty lag array.\n");
  }

  /* Temporary arrays */
  std::vector<axis> taxes;
  taxes.push_back(axis(_nlag,0.0,1.0));
  _b0 = new hypercube_float(taxes,true);
  _d0 = new hypercube_float(taxes,true);

  /* Construct the array of blocks */
  std::vector<int> b,e;
  find_optimal_sizes(_nd,j,b,e);
  for(int i = 0; i < (int)b.size(); ++i) {
    _blocks.push_back(block1d(b[i],e[i]));
  }
}

void lint_fast::find_optimal_sizes(int nd, int j, std::vector<int> &b, std::vector<int> &e) {
  int space = nd;
  b.push_back(0);
  for(int k = 0; k < _nf-1; ++k) {
    int sample = ceilf(static_cast<float>(space)/static_cast<float>(_nf-k-1));
    e.push_back(sample + b[k] - 1);
    if(k != _nf-2) b.push_back(e[k] + 1);
    space -= sample;
  }
  /* Take care of the endpoint */
  e[_nf-2] -= _lag.back();
}

std::vector<int> lint_fast::get_numfilt(){
  std::vector<int> fsize;
  fsize.push_back(_nlag);
  fsize.push_back(_nf);
  return fsize;
}

void lint_fast::print_blockinfo() {
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    fprintf(stderr,"blocknum = %d b1=%d e1=%d\n",ib,_blocks[ib].b1,_blocks[ib].e1);
  }
}

void lint_fast::print_laginfo() {
  for(int il = 0; il < _nlag; ++il) {
    fprintf(stderr,"%d lag=%d\n",il,_lag[il]);
  }
}

bool lint_fast::forward(bool add, my_vector *mod, my_vector *dat) {
  if(!add && dat) dat->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  int nd = d->get_axis(2).n;

  if(nd != _nd) {
    seperr("lv_conv1df: data axes must be same in operator as in constructor.\n");
  }

  if(m->get_axis(1).n != _nlag || m->get_axis(2).n != _nf) {
    seperr("lv_conv1df: input filter (model) is not same size as constructor.\n");
  }

  /* Loop over each block */
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    /* Get the filter coefficients at block corners */
    memcpy(_b0->vals,&m->vals[(ib+0)*_nlag],sizeof(float)*_nlag);
    memcpy(_d0->vals,&m->vals[(ib+1)*_nlag],sizeof(float)*_nlag);
    /* Compute interpolating factor to be added */
    _d0->scale_add(1.0,_b0,-1.0);
    double dist = 1.0/(_blocks[ib].e1 - _blocks[ib].b1 + 1.0);
    _d0->scale(dist);
    /* Loop over each point in the block */
    for(int id = _blocks[ib].b1; id <= _blocks[ib].e1; ++id) {
      _b0->scale_add(1.0,_d0,1.0);
      memcpy(&d->vals[id*_nlag],_b0->vals,sizeof(float)*_nlag);
    }
  }
  return true;
}

bool lint_fast::adjoint(bool add, my_vector *mod, my_vector *dat) {
  if(!add && mod) mod->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  int nd = d->get_axis(1).n;

  if(nd != _nd) {
    seperr("lv_conv1df: data axes must be same in operator as in constructor.\n");
  }

  if(m->get_axis(1).n != _nlag || m->get_axis(2).n != _nf) {
    seperr("lv_conv1df: input filter (model) is not same size as constructor.\n");
  }

  //TODO: Still need to finish the adjoint
  /* Loop over each block */
  double dist = 0.0;
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    /* Loop over each point in the block */
    _b0->zero(); _d0->zero();
    for(int id = _blocks[ib].e1; id >= _blocks[ib].b1; --id) {
      for(int il = 0; il < _nlag; ++il) {
        _d0->vals[il] += _b0->vals[il];
        _b0->vals[il] += d->vals[id];
      }
    }
    /* Adjoint linear interpolation */
    dist = 1.0/(_blocks[ib].e1 - _blocks[ib].b1 + 1.0);
    for(int il = 0; il < _nlag; ++il) {
      m->vals[(ib+0)*_nlag + il] += -dist*_d0->vals[il] + _b0->vals[il];
      m->vals[(ib+1)*_nlag + il] +=  dist*_d0->vals[il];
    }
  }
  return true;
}

