/*
 * lv_conv2df.cpp
 *
 *  Created on: Mar 17, 2017
 *      Author: joe
 */

#include <algorithm>
#include <tuple>
#include <math.h>
#include "lv_conv1dfmc.h"

lv_conv1dfmc::lv_conv1dfmc(int &nd, int &j, std::vector<int> nlag, hypercube_float *aux) {
  _nd = nd;
  _nf = static_cast<int>(ceilf(static_cast<float>(_nd-2.0)/static_cast<float>(j)) + 1);

  /* Default lag setting */
  int lag1min;
  /* Put the center lag at the middle of the lag 1 axis
   * this makes the min lag -n1lag/2 and then max lag
   * nlag1/2-1 for even lengths.
   * For odd lengths, minlag = -(n1lag-1)/2 and then
   * maxlag is -minlag
   */
  if(nlag[0]%2 == 0) {
    lag1min = -nlag[0]/2;
    //lag1max =  nlag[0]/2 - 1;
  } else {
    lag1min = -(nlag[0] - 1)/2;
    //lag1max = -lag1min;
  }

  for(int j = 0; j < nlag[1]; ++j) {
    for(int i = 0; i < nlag[0]; ++i) {
      if(lag1min + i < 0 && j == 0) {
        continue;
      } else {
        _lag.push_back(std::pair<int,int>(lag1min + i, j));
      }
    }
  }

  _nlag = (int)_lag.size();

  /* Temporary arrays */
  std::vector<axis> taxes;
  taxes.push_back(axis(_nlag,0.0,1.0));
  _b00 = new hypercube_float(taxes,true);
  _d01 = new hypercube_float(taxes,true);

  /* Find the min and max lag along axis 1 */
  decltype(_lag)::iterator maxlag1, minlag1;
  std::tie(minlag1,maxlag1)  = std::minmax_element(begin(_lag),end(_lag),
      [](const std::pair<int,int> &lag1, const std::pair<int,int> &lag2) {
    return lag1.first < lag2.first;
  });

  /* Find the min and max lag along axis 2 */
  decltype(_lag)::iterator maxlag2, minlag2;
  std::tie(minlag2,maxlag2)  = std::minmax_element(begin(_lag),end(_lag),
      [](const std::pair<int,int> &lag1, const std::pair<int,int> &lag2) {
    return lag1.second < lag2.second;
  });
  _maxlag2 = maxlag2->second;

  /* Construct the array of blocks */
  std::vector<int> b1, e1;
  find_optimal_sizes(_nd,j,_nf,b1,e1);
  for(int i = 0; i < (int)b1.size(); ++i) {
    /* Check the beginning and ending points */
    if(e1[i] == _nd-1) {
      e1[i] -= maxlag1->first;
    } if(b1[i] == 0) {
      b1[i] -= minlag1->first;
    }
    _blocks.push_back(block1d(b1[i],e1[i]));
  }

  /* Get the auxiliary */
  //TODO: What are some checks for the auxiliary?
  _aux = aux;
}

lv_conv1dfmc::lv_conv1dfmc(int &nd, int &j, std::vector<std::pair<int,int>> &lag, hypercube_float *aux) {
  _nd = nd;
  _nf = static_cast<int>(ceilf(static_cast<float>(nd-2.0)/static_cast<float>(j)) + 1);

  //TODO: Check the values of lags
  if(!lag.empty()) {
    _lag = lag;
  } else {
    seperr("lv_conv1dfmc: Input lag array is empty. Please initialize with a "
        "non-empty lag array.\n");
  }
  _nlag = (int)_lag.size();

  /* Temporary arrays */
  std::vector<axis> taxes;
  taxes.push_back(axis(_lag.size(),0.0,1.0));
  _b00  = new hypercube_float(taxes,true);
  _d01  = new hypercube_float(taxes,true);

  /* Find the min and max lag along axis 1 */
  decltype(_lag)::iterator maxlag1, minlag1;
  std::tie(minlag1,maxlag1)  = std::minmax_element(begin(_lag),end(_lag),
      [](const std::pair<int,int> &lag1, const std::pair<int,int> &lag2) {
    return lag1.first < lag2.first;
  });

  /* Find the min and max lag along axis 2 */
  decltype(_lag)::iterator maxlag2, minlag2;
  std::tie(minlag2,maxlag2)  = std::minmax_element(begin(_lag),end(_lag),
      [](const std::pair<int,int> &lag1, const std::pair<int,int> &lag2) {
    return lag1.second < lag2.second;
  });
  _maxlag2 = maxlag2->second;

  /* Construct the array of blocks */
  std::vector<int> b1, e1;
  find_optimal_sizes(_nd,j,_nf,b1,e1);
  for(int i = 0; i < (int)b1.size(); ++i) {
    /* Check the beginning and ending points */
    if(e1[i] == _nd-1) {
      e1[i] -= maxlag1->first;
    } if(b1[i] == 0) {
      b1[i] -= minlag1->first;
    }
    _blocks.push_back(block1d(b1[i],e1[i]));
  }

  /* Get the auxiliary */
  //TODO: What are some checks for the auxiliary?
  _aux = aux;
}

void lv_conv1dfmc::find_optimal_sizes(int nd, int j, int nf, std::vector<int> &b, std::vector<int> &e) {
  int space = nd;
  b.push_back(0);
  for(int k = 0; k < nf-1; ++k) {
    int sample = ceilf(static_cast<float>(space)/static_cast<float>(nf-k-1));
    e.push_back(sample + b[k] - 1);
    if(k != nf-2) b.push_back(e[k] + 1);
    space -= sample;
  }
}

//TODO: Change so that it returns a std::vector<axis>
std::vector<int> lv_conv1dfmc::get_numfilt() {
  std::vector<int> fsize;
  fsize.push_back(_nlag);
  fsize.push_back(_nf);
  return fsize;
}

void lv_conv1dfmc::set_zerolag(hypercube_float *flt) {
  for(int if1 = 0; if1 < _nf; ++if1) {
    flt->vals[if1*_nlag] = 1.0;
  }
}

void lv_conv1dfmc::print_blockinfo() {
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    fprintf(stderr,"blocknum=%d b1=%d e1=%d\n",ib,_blocks[ib].b1,_blocks[ib].e1);
  }
}

void lv_conv1dfmc::print_laginfo() {
  for(int il = 0; il < _nlag; ++il) {
    fprintf(stderr,"%d lag1=%d lag2=%d\n",il,_lag[il].first,_lag[il].second);
  }
}

bool lv_conv1dfmc::forward(bool add, my_vector *mod, my_vector *dat) {
  if(!add && dat) dat->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  int nd1 = d->get_axis(1).n;
  int nc  = d->get_axis(2).n;

  if(nd1 != _nd) {
    seperr("lv_conv1dfmc: data axes must be same in operator as in constructor.\n");
  }

  if(nc != (_maxlag2+1)) {
    seperr("lv_conv1dfmc: number of filter coefficients along axis 2 must equal number of components.\n");
  }

  if(!d->check_match(_aux)) {
    seperr("lv_conv1dfmc: input data and auxiliary should be same size.\n");
  }

  if(m->get_axis(1).n != _nlag || m->get_axis(2).n != _nf) {
    seperr("lv_conv2df: input filter (model) is not same size as constructor.\n");
  }

  double dist = 0.0;
  /* Loop over each block */
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    /* Get the filter coefficients at block corners */
    memcpy(_b00->vals,&m->vals[(ib+0)*_nlag],sizeof(float)*_nlag);
    memcpy(_d01->vals,&m->vals[(ib+1)*_nlag],sizeof(float)*_nlag);
    dist = 1.0/(_blocks[ib].e1 - _blocks[ib].b1 + 1.0);
    /* Compute axis 1 interpolating factor to be added */
    _d01->scale_add(1.0,_b00,-1.0);
    _d01->scale(dist);
    for(int id1 = _blocks[ib].b1; id1 <= _blocks[ib].e1; ++id1) {
      /* Perform the convolution with interpolated filter */
      for(int il = 0; il < _nlag; ++il) {
        d->vals[(_lag[il].second)*nd1 + id1+_lag[il].first] += _b00->vals[il]*_aux->vals[id1];
      }
      /* Interpolate along axis 1 */
      _b00->scale_add(1.0,_d01,1.0);
    }
  }
  return true;
}

void lv_conv1dfmc::set_data(hypercube_float *data) {
  /* Create a filter with only zero-lag coefficients set */
  std::vector<axis> faxes;
  faxes.push_back(axis(_nlag ,0.0,1.0));
  faxes.push_back(axis(_nf,0.0,1.0));
  hypercube_float *tflt = new hypercube_float(faxes,true);
  tflt->zero();
  set_zerolag(tflt);
  /* Convolve auxiliary with this temporary filter */
  forward(false,tflt,data);
  delete tflt;
}

bool lv_conv1dfmc::adjoint(bool add, my_vector *mod, my_vector *dat) {
  if(!add && mod) mod->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  int nd1 = d->get_axis(1).n;
  int nc  = d->get_axis(2).n;

  if(nd1 != _nd) {
    seperr("lv_conv1dfmc: data axes must be same in operator as in constructor.\n");
  }

  if(nc != (_maxlag2+1)) {
    seperr("lv_conv1dfmc: number of filter coefficients along axis 2 must equal number of components.\n");
  }

  if(!d->check_match(_aux)) {
    seperr("lv_conv1dfmc: input data and auxiliary should be same size.\n");
  }

  if(m->get_axis(1).n != _nlag || m->get_axis(2).n != _nf) {
    seperr("lv_conv1dfmc: input filter (model) is not same size as constructor.\n");
  }

  double dist = 0.0,x = 0.0;
  /* Loop over blocks */
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    /* Loop over each point in the block */
    _b00->zero(); _d01->zero();
    for(int id1 = _blocks[ib].e1; id1 >= _blocks[ib].b1; --id1) {
      for(int il = 0; il < _nlag; ++il) {
        x = _aux->vals[id1] * d->vals[(_lag[il].second)*nd1 + id1+_lag[il].first];
        _d01->vals[il] += _b00->vals[il];
        _b00->vals[il] += x;
      }
    }
    dist = 1.0/(_blocks[ib].e1 - _blocks[ib].b1 + 1.0);
    for(int il = 0; il < _nlag; ++il) {
      m->vals[(ib+0)*_nlag + il] += -dist*_d01->vals[il] + _b00->vals[il];
      m->vals[(ib+1)*_nlag + il] +=  dist*_d01->vals[il];
    }
  }
  return true;
}
