/*
 * lv_conv1dm.cpp
 *
 *  Created on: Mar 15, 2017
 *      Author: joe
 */

#include "lv_conv1dm.h"
#include <math.h>

lv_conv1dm::lv_conv1dm(int nd, int j, int nlag, hypercube_float *aux) {

  /* Compute the number of filters on each axis */
  _nd = nd;
  _nf = static_cast<int>(ceilf(static_cast<float>(_nd-2.0)/static_cast<float>(j)) + 1);
  _nlag = nlag;

  /* Assume lags increase linearly */
  for(int i = 0; i < _nlag; ++i) {
    _lag.push_back(i);
  }

  /* Temporary arrays */
  std::vector<axis> taxes;
  taxes.push_back(axis(_nlag,0.0,1.0));
  _b0 = new hypercube_float(taxes,true);
  _d0 = new hypercube_float(taxes,true);

  /* Construct the array of blocks */
  std::vector<int> b, e;
  find_optimal_sizes(_nd,j,_nf,b,e);
  for(int i = 0; i < (int)b.size(); ++i) {
    _blocks.push_back(block1d(b[i],e[i]));
  }

  /* Get the auxiliary (filter) */
  if(aux->get_axis(1).n != _nlag || aux->get_axis(2).n != _nf) {
    seperr("Calculated filter lengths %d %d must be same length as input filter %d %d \n",_nlag,_nf,
        aux->get_axis(1).n,aux->get_axis(2).n);
  }
  _aux = aux;
}

lv_conv1dm::lv_conv1dm(int nd, int j, std::vector<int> lag, hypercube_float *aux) {
  /* Compute the number of filters on each axis */
  _nd = nd;
  _nf = static_cast<int>(ceilf(static_cast<float>(_nd-2.0)/static_cast<float>(j)) + 1);
  _nlag = (int)lag.size();

  /* Save lags */
  if(!lag.empty()) {
    _lag = lag;
  } else {
    seperr("lv_conv1df: Input lag array is empty. Please initialize with a "
        "non-empty lag array.\n");
  }

  /* Temporary arrays */
  std::vector<axis> taxes;
  taxes.push_back(axis(_nlag,0.0,1.0));
  _b0 = new hypercube_float(taxes,true);
  _d0 = new hypercube_float(taxes,true);

  /* Construct the array of blocks */
  std::vector<int> b,e;
  find_optimal_sizes(_nd,j,_nf,b,e);
  for(int i = 0; i < (int)b.size(); ++i) {
    _blocks.push_back(block1d(b[i],e[i]));
  }

  /* Get the auxiliary (filter) */
  if(aux->get_axis(1).n != _nlag || aux->get_axis(2).n != _nf) {
    seperr("Calculated filter lengths %d %d must be same length as input filter %d %d \n",_nlag,_nf,
        aux->get_axis(1).n,aux->get_axis(2).n);
  }
  _aux = aux;
}

void lv_conv1dm::find_optimal_sizes(int nd, int j, int nf, std::vector<int> &b, std::vector<int> &e) {
  int space = nd;
  b.push_back(0);
  for(int k = 0; k < nf-1; ++k) {
    int sample = ceilf(static_cast<float>(space)/static_cast<float>(nf-k-1));
    e.push_back(sample + b[k] - 1);
    if(k != nf-2) b.push_back(e[k] + 1);
    space -= sample;
  }
  /* Take care of the endpoint */
  e[nf-2] -= _lag.back();
}

std::vector<int> lv_conv1dm::get_numfilt(){
  std::vector<int> fsize;
  fsize.push_back(_nlag);
  fsize.push_back(_nf);
  return fsize;
}

void lv_conv1dm::set_zerolag(hypercube_float *flt) {
  for(int if1 = 0; if1 < _nf; ++if1) {
    flt->vals[if1*_nlag] = 1.0;
  }
}

void lv_conv1dm::print_blockinfo() {
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    fprintf(stderr,"blocknum = %d b1=%d e1=%d\n",ib,_blocks[ib].b1,_blocks[ib].e1);
  }
}

void lv_conv1dm::print_laginfo() {
  for(int il = 0; il < _nlag; ++il) {
    fprintf(stderr,"%d lag=%d\n",il,_lag[il]);
  }
}

bool lv_conv1dm::forward(bool add, my_vector *mod, my_vector *dat) {
  if(!add && dat) dat->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  int nd = d->get_axis(1).n;

  if(nd != _nd) {
    seperr("lv_conv1dm: data axes must be same in operator as in constructor.\n");
  }

  if(!d->check_match(m)) {
    seperr("lv_conv1dm: input data and auxiliary should be same size.\n");
  }

  /* Loop over each block */
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    /* Get the filter coefficients at block corners */
    memcpy(_b0->vals,&_aux->vals[(ib+0)*_nlag],sizeof(float)*_nlag);
    memcpy(_d0->vals,&_aux->vals[(ib+1)*_nlag],sizeof(float)*_nlag);
    /* Compute interpolating factor to be added */
    _d0->scale_add(1.0,_b0,-1.0);
    double dist = 1.0/(_blocks[ib].e1 - _blocks[ib].b1 + 1.0);
    _d0->scale(dist);
    /* Loop over each point in the block */
    for(int id = _blocks[ib].b1; id <= _blocks[ib].e1; ++id) {
      /* Convolve with interpolated filter */
      for(int il = 0; il < _nlag; ++il) {
        d->vals[id + _lag[il]] += scale * _b0->vals[il] * m->vals[id];
      }
      _b0->scale_add(1.0,_d0,1.0);
    }
  }
  return true;
}

bool lv_conv1dm::adjoint(bool add, my_vector *mod, my_vector *dat) {
  if(!add && mod) mod->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  int nd = d->get_axis(1).n;

  if(nd != _nd) {
    seperr("lv_conv1dm: data axes must be same in operator as in constructor.\n");
  }

  if(!d->check_match(m)) {
    seperr("lv_conv1dm: input data and auxiliary should be same size.\n");
  }

  /* Loop over each block */
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    /* Get the filter coefficients at block corners */
    memcpy(_b0->vals,&_aux->vals[(ib+0)*_nlag],sizeof(float)*_nlag);
    memcpy(_d0->vals,&_aux->vals[(ib+1)*_nlag],sizeof(float)*_nlag);
    /* Compute interpolating factor to be added */
    _d0->scale_add(1.0,_b0,-1.0);
    double dist = 1.0/(_blocks[ib].e1 - _blocks[ib].b1 + 1.0);
    _d0->scale(dist);
    /* Loop over each point in the block */
    for(int id = _blocks[ib].b1; id <= _blocks[ib].e1; ++id) {
      /* Convolve with interpolated filter */
      for(int il = 0; il < _nlag; ++il) {
        m->vals[id] += scale * _b0->vals[il] * d->vals[id + _lag[il]];
      }
      _b0->scale_add(1.0,_d0,1.0);
    }
  }
  return true;
}
