/*
 * lv_conv2dm.cpp
 *
 *  Created on: Mar 23, 2017
 *      Author: joe
 */

#include <algorithm>
#include <tuple>
#include <math.h>
#include "lv_conv2dm.h"

lv_conv2dm::lv_conv2dm(std::vector<int> nd, std::vector<int> &j,
    std::vector<int> nlag, hypercube_float *aux) {
  _nd = nd;
  _nf.push_back(static_cast<int>(ceilf(static_cast<float>(_nd[0]-2.0)/static_cast<float>(j[0])) + 1));
  _nf.push_back(static_cast<int>(ceilf(static_cast<float>(_nd[1]-2.0)/static_cast<float>(j[1])) + 1));

  /* Default lag setting */
  int lag1min;
  /* Put the center lag at the middle of the lag 1 axis
   * this makes the min lag -n1lag/2 and then max lag
   * nlag1/2-1 for even lengths.
   * For odd lengths, minlag = -(n1lag-1)/2 and then
   * maxlag is -minlag
   */
  if(nlag[0]%2 == 0) {
    lag1min = -nlag[0]/2;
    //lag1max =  nlag[0]/2 - 1;
  } else {
    lag1min = -(nlag[0] - 1)/2;
    //lag1max = -lag1min;
  }

  for(int j = 0; j < nlag[1]; ++j) {
    for(int i = 0; i < nlag[0]; ++i) {
      if(lag1min + i < 0 && j == 0) {
        continue;
      } else {
        _lag.push_back(std::pair<int,int>(lag1min + i, j));
      }
    }
  }

  _nlag = (int)_lag.size();

  /* Temporary arrays */
  std::vector<axis> taxes;
  taxes.push_back(axis(_nlag,0.0,1.0));
  _b00 = new hypercube_float(taxes,true);
  _b01 = new hypercube_float(taxes,true);
  _d01 = new hypercube_float(taxes,true);
  _d02 = new hypercube_float(taxes,true);
  _d12 = new hypercube_float(taxes,true);
  _bin = new hypercube_float(taxes,true);

  /* Find the min and max lag along axis 1 */
  decltype(_lag)::iterator maxlag1, minlag1;
  std::tie(minlag1,maxlag1)  = std::minmax_element(begin(_lag),end(_lag),
      [](const std::pair<int,int> &lag1, const std::pair<int,int> &lag2) {
    return lag1.first < lag2.first;
  });

  /* Find the min and max lag along axis 2 */
  decltype(_lag)::iterator maxlag2, minlag2;
  std::tie(minlag2,maxlag2)  = std::minmax_element(begin(_lag),end(_lag),
      [](const std::pair<int,int> &lag1, const std::pair<int,int> &lag2) {
    return lag1.second < lag2.second;
  });

  /* Construct the array of blocks */
  std::vector<int> b1, e1, b2, e2;
  find_optimal_sizes(_nd[0],j[0],_nf[0],b1,e1);
  find_optimal_sizes(_nd[1],j[1],_nf[1],b2,e2);
  for(int j = 0; j < (int)b2.size(); ++j) {
    for(int i = 0; i < (int)b1.size(); ++i) {
      /* Check the beginning and ending points */
      if(e1[i] == _nd[0]-1) {
        e1[i] -= maxlag1->first;
      } if(e2[j] == _nd[1]-1) {
        e2[j] -= maxlag2->second;
      } if(b1[i] == 0) {
        b1[i] -= minlag1->first;
      } if(b2[j] == 0) {
        b2[j] -= minlag2->second;
      }
      _blocks.push_back(block2d(b1[i],e1[i],b2[j],e2[j]));
    }
  }

  /* Get the auxiliary */
  if(aux->get_axis(1).n != _nlag || aux->get_axis(2).n != _nf[0] || aux->get_axis(3).n != _nf[1]) {
    seperr("Calculated filter lengths %d %d %d %d must be same length as input filter %d %d %d %d \n",_nlag,_nf[0],_nf[1],
        aux->get_axis(1).n,aux->get_axis(2).n,aux->get_axis(3).n);
  }
  _aux = aux;
}

lv_conv2dm::lv_conv2dm(std::vector<int> nd, std::vector<int> &j,
    std::vector<std::pair<int,int>> &lag, hypercube_float *aux) {
  _nd = nd;
  _nf.push_back(static_cast<int>(ceilf(static_cast<float>(nd[0]-2.0)/static_cast<float>(j[0])) + 1));
  _nf.push_back(static_cast<int>(ceilf(static_cast<float>(nd[1]-2.0)/static_cast<float>(j[1])) + 1));

  //TODO: Check the values of lags
  _lag = lag;
  _nlag = (int)_lag.size();

  /* Temporary arrays */
  std::vector<axis> taxes;
  taxes.push_back(axis(_lag.size(),0.0,1.0));
  _b00  = new hypercube_float(taxes,true);
  _b01  = new hypercube_float(taxes,true);
  _d01  = new hypercube_float(taxes,true);
  _d02 = new hypercube_float(taxes,true);
  _d12 = new hypercube_float(taxes,true);
  _bin = new hypercube_float(taxes,true);

  /* Find the min and max lag along axis 1 */
  decltype(_lag)::iterator maxlag1, minlag1;
  std::tie(minlag1,maxlag1)  = std::minmax_element(begin(_lag),end(_lag),
      [](const std::pair<int,int> &lag1, const std::pair<int,int> &lag2) {
    return lag1.first < lag2.first;
  });

  /* Find the min and max lag along axis 2 */
  decltype(_lag)::iterator maxlag2, minlag2;
  std::tie(minlag2,maxlag2)  = std::minmax_element(begin(_lag),end(_lag),
      [](const std::pair<int,int> &lag1, const std::pair<int,int> &lag2) {
    return lag1.second < lag2.second;
  });

  /* Construct the array of blocks */
  std::vector<int> b1, e1, b2, e2;
  find_optimal_sizes(_nd[0],j[0],_nf[0],b1,e1);
  find_optimal_sizes(_nd[1],j[1],_nf[1],b2,e2);
  for(int j = 0; j < (int)b2.size(); ++j) {
    for(int i = 0; i < (int)b1.size(); ++i) {
      /* Check the beginning and ending points */
      if(e1[i] == _nd[0]-1) {
        e1[i] -= maxlag1->first;
      } if(e2[j] == _nd[1]-1) {
        e2[j] -= maxlag2->second;
      } if(b1[i] == 0) {
        b1[i] -= minlag1->first;
      } if(b2[j] == 0) {
        b2[j] -= minlag2->second;
      }
      _blocks.push_back(block2d(b1[i],e1[i],b2[j],e2[j]));
    }
  }

  /* Get the auxiliary */
  if(aux->get_axis(1).n != _nlag || aux->get_axis(2).n != _nf[0] || aux->get_axis(3).n != _nf[1]) {
    seperr("Calculate filter lengths must be same length as input filter\n");
  }
  _aux = aux;
}

void lv_conv2dm::find_optimal_sizes(int nd, int j, int nf, std::vector<int> &b, std::vector<int> &e) {
  int space = nd;
  b.push_back(0);
  for(int k = 0; k < nf-1; ++k) {
    int sample = ceilf(static_cast<float>(space)/static_cast<float>(nf-k-1));
    e.push_back(sample + b[k] - 1);
    if(k != nf-2) b.push_back(e[k] + 1);
    space -= sample;
  }
}

//TODO: Change so that it returns a std::vector<axis>
std::vector<int> lv_conv2dm::get_numfilt() {
  std::vector<int> fsize;
  fsize.push_back(_nlag);
  fsize.push_back(_nf[0]);
  fsize.push_back(_nf[1]);
  return fsize;
}

void lv_conv2dm::set_zerolag(hypercube_float *flt) {
  for(int if2 = 0; if2 < _nf[1]; ++if2) {
    for(int if1 = 0; if1 < _nf[0]; ++if1) {
      flt->vals[if2*_nf[0]*_nlag + if1*_nlag] = 1.0;
    }
  }
}

void lv_conv2dm::print_blockinfo() {
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    fprintf(stderr,"blocknum=%d b1=%d e1=%d b2=%d e2=%d\n",ib,_blocks[ib].b1,_blocks[ib].e1,_blocks[ib].b2,_blocks[ib].e2);
  }
}

void lv_conv2dm::print_laginfo() {
  for(int il = 0; il < _nlag; ++il) {
    fprintf(stderr,"%d lag1=%d lag2=%d\n",il,_lag[il].first,_lag[il].second);
  }
}

bool lv_conv2dm::forward(bool add, my_vector *mod, my_vector *dat) {
  if(!add && dat) dat->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  int nd1 = d->get_axis(1).n;
  int nd2 = d->get_axis(2).n;

  if(nd1 != _nd[0] || nd2 != _nd[1]) {
    seperr("lv_conv2df: data axes must be same in operator as in constructor.\n");
  }

  if(!d->check_match(m)) {
    seperr("lv_conv2df: input data and auxiliary should be same size.\n");
  }

  double dist1 = 0.0, dist2 = 0.0;
  /* Loop over each block */
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    /* Get the locations of the block corners */
    int k1 = ib%(_nf[0]-1);
    int k2 = ib/(_nf[0]-1);
    /* Get the filter coefficients at block corners */
    memcpy(_b00->vals,&_aux->vals[((k2+0)*_nf[0] + k1+0)*_nlag],sizeof(float)*_nlag);
    memcpy(_b01->vals,&_aux->vals[((k2+0)*_nf[0] + k1+1)*_nlag],sizeof(float)*_nlag);
    memcpy(_d02->vals,&_aux->vals[((k2+1)*_nf[0] + k1+0)*_nlag],sizeof(float)*_nlag);
    memcpy(_d12->vals,&_aux->vals[((k2+1)*_nf[0] + k1+1)*_nlag],sizeof(float)*_nlag);
    /* Compute axis 2 interpolating factors to be added */
    _d02->scale_add(1.0,_b00,-1.0); _d12->scale_add(1.0,_b01,-1.0);
    dist2 = 1.0/(_blocks[ib].e2 - _blocks[ib].b2 + 1.0);
    dist1 = 1.0/(_blocks[ib].e1 - _blocks[ib].b1 + 1.0);
    _d02->scale(dist2); _d12->scale(dist2);
    /* Loop over each point in block */
    for(int id2 = _blocks[ib].b2; id2 <= _blocks[ib].e2; ++id2) {
      memcpy(_bin->vals,_b00->vals,sizeof(float)*_nlag);
      memcpy(_d01->vals,_b01->vals,sizeof(float)*_nlag);
      /* Compute axis 1 interpolating factor to be added */
      _d01->scale_add(1.0,_b00,-1.0);
      _d01->scale(dist1);
      for(int id1 = _blocks[ib].b1; id1 <= _blocks[ib].e1; ++id1) {
        /* Perform the convolution with interpolated filter */
        for(int il = 0; il < _nlag; ++il) {
          d->vals[(id2+_lag[il].second)*nd1 + id1+_lag[il].first] += scale * _bin->vals[il]*m->vals[id2*nd1 + id1];
        }
        _bin->scale_add(1.0,_d01,1.0);
      }
      _b00->scale_add(1.0,_d02,1.0);
      _b01->scale_add(1.0,_d12,1.0);
    }
  }
  return true;
}

void lv_conv2dm::set_data(hypercube_float *data) {
  /* Create a filter with only zero-lag coefficients set */
  std::vector<axis> faxes;
  faxes.push_back(axis(_nlag ,0.0,1.0));
  faxes.push_back(axis(_nf[0],0.0,1.0));
  faxes.push_back(axis(_nf[1],0.0,1.0));
  hypercube_float *tflt = new hypercube_float(faxes,true);
  set_zerolag(tflt);
  /* Convolve auxiliary with this temporary filter */
  forward(false,tflt,data);
  delete tflt;
}

bool lv_conv2dm::adjoint(bool add, my_vector *mod, my_vector *dat) {
  if(!add && mod) mod->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  int nd1 = d->get_axis(1).n;
  int nd2 = d->get_axis(2).n;

  if(nd1 != _nd[0] || nd2 != _nd[1]) {
    seperr("lv_conv2dm: data axes must be same in operator as in constructor.\n");
  }

  if(!d->check_match(m)) {
    seperr("lv_conv2dm: input data and model should be same size.\n");
  }


  double dist1 = 0.0, dist2 = 0.0;
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    /* Get the locations of the block corners */
    int k1 = ib%(_nf[0]-1);
    int k2 = ib/(_nf[0]-1);
    /* Get the filter coefficients at block corners */
    memcpy(_b00->vals,&_aux->vals[((k2+0)*_nf[0] + k1+0)*_nlag],sizeof(float)*_nlag);
    memcpy(_b01->vals,&_aux->vals[((k2+0)*_nf[0] + k1+1)*_nlag],sizeof(float)*_nlag);
    memcpy(_d02->vals,&_aux->vals[((k2+1)*_nf[0] + k1+0)*_nlag],sizeof(float)*_nlag);
    memcpy(_d12->vals,&_aux->vals[((k2+1)*_nf[0] + k1+1)*_nlag],sizeof(float)*_nlag);
    /* Compute axis 2 interpolating factors to be added */
    _d02->scale_add(1.0,_b00,-1.0); _d12->scale_add(1.0,_b01,-1.0);
    dist2 = 1.0/(_blocks[ib].e2 - _blocks[ib].b2 + 1.0);
    dist1 = 1.0/(_blocks[ib].e1 - _blocks[ib].b1 + 1.0);
    _d02->scale(dist2); _d12->scale(dist2);
    /* Loop over each point in block */
    for(int id2 = _blocks[ib].b2; id2 <= _blocks[ib].e2; ++id2) {
      memcpy(_bin->vals,_b00->vals,sizeof(float)*_nlag);
      memcpy(_d01->vals,_b01->vals,sizeof(float)*_nlag);
      /* Compute axis 1 interpolating factor to be added */
      _d01->scale_add(1.0,_b00,-1.0);
      _d01->scale(dist1);
      for(int id1 = _blocks[ib].b1; id1 <= _blocks[ib].e1; ++id1) {
        /* Perform the convolution with interpolated filter */
        for(int il = 0; il < _nlag; ++il) {
          m->vals[id2*nd1 + id1] += scale * _bin->vals[il]*d->vals[(id2+_lag[il].second)*nd1 + id1+_lag[il].first];
        }
        _bin->scale_add(1.0,_d01,1.0);
      }
      _b00->scale_add(1.0,_d02,1.0);
      _b01->scale_add(1.0,_d12,1.0);
    }
  }
  return true;
}
