/*
 * lv_conv3df.cpp
 *
 *  Created on: Mar 21, 2017
 *      Author: joe
 */

#include <algorithm>
#include <tuple>
#include "lv_conv3df.h"
#include "math.h"

lv_conv3df::lv_conv3df(std::vector<int> nd, std::vector<int> &j,
    std::vector<int> nlag, hypercube_float *aux) {
  _nd = nd;
  _nf.push_back(static_cast<int>(ceilf(static_cast<float>(_nd[0]-2.0)/static_cast<float>(j[0])) + 1));
  _nf.push_back(static_cast<int>(ceilf(static_cast<float>(_nd[1]-2.0)/static_cast<float>(j[1])) + 1));
  _nf.push_back(static_cast<int>(ceilf(static_cast<float>(_nd[2]-2.0)/static_cast<float>(j[2])) + 1));

  /* Default lag setting */
  int lag1min;
  /* Put the center lag at the middle of the lag 1 axis
   * this makes the min lag -n1lag/2 and then max lag
   * nlag1/2-1 for even lengths.
   * For odd lengths, minlag = -(n1lag-1)/2 and then
   * maxlag is -minlag
   */
  if(nlag[0]%2 == 0) {
    lag1min = -nlag[0]/2;
    //lag1max =  nlag[0]/2 - 1;
  } else {
    lag1min = -(nlag[0] - 1)/2;
    //lag1max = -lag1min;
  }

  for(int k = 0; k < nlag[2]; ++k) {
    for(int j = 0; j < nlag[1]; ++j) {
      for(int i = 0; i < nlag[0]; ++i) {
        if(lag1min + i < 0 && j == 0 && k == 0) {
          continue;
        } else {
          _lag.push_back(triple(lag1min + i, j, k));
        }
      }
    }
  }

  _nlag = (int)_lag.size();

  /* Temporary arrays */
  std::vector<axis> taxes;
  taxes.push_back(axis(_nlag,0.0,1.0));
  _b000 = new hypercube_float(taxes,true);
  _b100 = new hypercube_float(taxes,true);
  _b010 = new hypercube_float(taxes,true);
  _b110 = new hypercube_float(taxes,true);
  _d001 = new hypercube_float(taxes,true);
  _d011 = new hypercube_float(taxes,true);
  _d101 = new hypercube_float(taxes,true);
  _d111 = new hypercube_float(taxes,true);
  _b00  = new hypercube_float(taxes,true);
  _b10  = new hypercube_float(taxes,true);
  _d00  = new hypercube_float(taxes,true);
  _d01  = new hypercube_float(taxes,true);
  _d0  = new hypercube_float(taxes,true);
  _bin  = new hypercube_float(taxes,true);

  /* Find the min and max lag along axis 1 */
  decltype(_lag)::iterator maxlag1, minlag1;
  std::tie(minlag1,maxlag1)  = std::minmax_element(begin(_lag),end(_lag),
      [](const triple &lag1, const triple &lag2) {
    return lag1.first < lag2.first;
  });

  /* Find the min and max lag along axis 2 */
  decltype(_lag)::iterator maxlag2, minlag2;
  std::tie(minlag2,maxlag2)  = std::minmax_element(begin(_lag),end(_lag),
      [](const triple &lag1, const triple &lag2) {
    return lag1.second < lag2.second;
  });

  /* Find the min and max lag along axis 3 */
  decltype(_lag)::iterator maxlag3, minlag3;
  std::tie(minlag3,maxlag3)  = std::minmax_element(begin(_lag),end(_lag),
      [](const triple &lag1, const triple &lag2) {
    return lag1.third < lag2.third;
  });

  /* Construct the array of blocks */
  std::vector<int> b1, e1, b2, e2, b3, e3;
  find_optimal_sizes(_nd[0],j[0],_nf[0],b1,e1);
  find_optimal_sizes(_nd[1],j[1],_nf[1],b2,e2);
  find_optimal_sizes(_nd[2],j[2],_nf[2],b3,e3);
  for(int k = 0; k < (int)b3.size(); ++k) {
    for(int j = 0; j < (int)b2.size(); ++j) {
      for(int i = 0; i < (int)b1.size(); ++i) {
        /* Check the beginning and ending points */
        if(e1[i] == _nd[0]-1) {
          e1[i] -= maxlag1->first;
        } if(e2[j] == _nd[1]-1) {
          e2[j] -= maxlag2->second;
        } if(e3[k] == _nd[2]-1) {
          e3[k] -= maxlag3->third;
        } if(b1[i] == 0) {
          b1[i] -= minlag1->first;
        } if(b2[j] == 0) {
          b2[j] -= minlag2->second;
        } if(b3[k] == 0) {
          b3[k] -= minlag3->third;
        }
        _blocks.push_back(block3d(b1[i],e1[i],b2[j],e2[j],b3[k],e3[k]));
      }
    }
  }

  /* Get the auxiliary */
  //TODO: What are some checks for the auxiliary?
  _aux = aux;
}
lv_conv3df::lv_conv3df(std::vector<int> nd, std::vector<int> &j,
    std::vector<triple> &lag, hypercube_float *aux) {
  _nd = nd;
  _nf.push_back(static_cast<int>(ceilf(static_cast<float>(_nd[0]-2.0)/static_cast<float>(j[0])) + 1));
  _nf.push_back(static_cast<int>(ceilf(static_cast<float>(_nd[1]-2.0)/static_cast<float>(j[1])) + 1));
  _nf.push_back(static_cast<int>(ceilf(static_cast<float>(_nd[2]-2.0)/static_cast<float>(j[2])) + 1));

  //TODO: Check the values of lag
  _lag = lag;
  _nlag = (int)_lag.size();

  /* Temporary arrays */
  std::vector<axis> taxes;
  taxes.push_back(axis(_nlag,0.0,1.0));
  /* Temporary arrays */
  _b000 = new hypercube_float(taxes,true);
  _b100 = new hypercube_float(taxes,true);
  _b010 = new hypercube_float(taxes,true);
  _b110 = new hypercube_float(taxes,true);
  _d001 = new hypercube_float(taxes,true);
  _d011 = new hypercube_float(taxes,true);
  _d101 = new hypercube_float(taxes,true);
  _d111 = new hypercube_float(taxes,true);
  _b00  = new hypercube_float(taxes,true);
  _b10  = new hypercube_float(taxes,true);
  _d00  = new hypercube_float(taxes,true);
  _d01  = new hypercube_float(taxes,true);
  _d0   = new hypercube_float(taxes,true);
  _bin  = new hypercube_float(taxes,true);

  /* Find the min and max lag along axis 1 */
  decltype(_lag)::iterator maxlag1, minlag1;
  std::tie(minlag1,maxlag1)  = std::minmax_element(begin(_lag),end(_lag),
      [](const triple &lag1, const triple &lag2) {
    return lag1.first < lag2.first;
  });

  /* Find the min and max lag along axis 2 */
  decltype(_lag)::iterator maxlag2, minlag2;
  std::tie(minlag2,maxlag2)  = std::minmax_element(begin(_lag),end(_lag),
      [](const triple &lag1, const triple &lag2) {
    return lag1.second < lag2.second;
  });

  /* Find the min and max lag along axis 3 */
  decltype(_lag)::iterator maxlag3, minlag3;
  std::tie(minlag3,maxlag3)  = std::minmax_element(begin(_lag),end(_lag),
      [](const triple &lag1, const triple &lag2) {
    return lag1.third < lag2.third;
  });

  /* Construct the array of blocks */
  std::vector<int> b1, e1, b2, e2, b3, e3;
  find_optimal_sizes(_nd[0],j[0],_nf[0],b1,e1);
  find_optimal_sizes(_nd[1],j[1],_nf[1],b2,e2);
  find_optimal_sizes(_nd[2],j[2],_nf[2],b3,e3);
  for(int k = 0; k < (int)b3.size(); ++k) {
    for(int j = 0; j < (int)b2.size(); ++j) {
      for(int i = 0; i < (int)b1.size(); ++i) {
        /* Check the beginning and ending points */
        if(e1[i] == _nd[0]-1) {
          e1[i] -= maxlag1->first;
        } if(e2[j] == _nd[1]-1) {
          e2[j] -= maxlag2->second;
        } if(e3[k] == _nd[2]-1) {
          e3[k] -= maxlag3->third;
        } if(b1[i] == 0) {
          b1[i] -= minlag1->first;
        } if(b2[j] == 0) {
          b2[j] -= minlag2->second;
        } if(b3[k] == 0) {
          b3[k] -= minlag3->third;
        }
        _blocks.push_back(block3d(b1[i],e1[i],b2[j],e2[j],b3[k],e3[k]));
      }
    }
  }

  /* Get the auxiliary */
  //TODO: What are some checks for the auxiliary?
  _aux = aux;
}

void lv_conv3df::find_optimal_sizes(int nd, int j, int nf, std::vector<int> &b, std::vector<int> &e) {
  int space = nd;
  b.push_back(0);
  for(int k = 0; k < nf-1; ++k) {
    int sample = ceilf(static_cast<float>(space)/static_cast<float>(nf-k-1));
    e.push_back(sample + b[k] - 1);
    if(k != nf-2) b.push_back(e[k] + 1);
    space -= sample;
  }
  if(nf-1 == 0) {
    e.push_back(0);
  }
}

std::vector<axis> lv_conv3df::get_filtaxes() {
  std::vector<axis> faxes;
  faxes.push_back(axis(_nlag ,0.0,1.0));
  faxes.push_back(axis(_nf[0],0.0,1.0));
  faxes.push_back(axis(_nf[1],0.0,1.0));
  faxes.push_back(axis(_nf[2],0.0,1.0));
  return faxes;
}

void lv_conv3df::set_zerolag(hypercube_float *flt) {
  for(int if3 = 0; if3 < _nf[2]; ++if3) {
    for(int if2 = 0; if2 < _nf[1]; ++if2) {
      for(int if1 = 0; if1 < _nf[0]; ++if1) {
        flt->vals[if3*_nf[1]*_nf[0]*_nlag + if2*_nf[0]*_nlag + if1*_nlag] = 1.0;
      }
    }
  }
}

void lv_conv3df::print_blockinfo() {
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    fprintf(stderr,"blocknum=%d b1=%d e1=%d b2=%d e2=%d b3=%d e3=%d\n",ib,_blocks[ib].b1,_blocks[ib].e1,
        _blocks[ib].b2,_blocks[ib].e2,_blocks[ib].b3,_blocks[ib].e3);
  }
}

void lv_conv3df::print_laginfo() {
  for(int il = 0; il < _nlag; ++il) {
    fprintf(stderr,"%d lag1=%d lag2=%d lag3=%d\n",il,_lag[il].first,_lag[il].second,_lag[il].third);
  }
}

bool lv_conv3df::forward(bool add, my_vector *mod, my_vector *dat) {
  if(!add && dat) dat->zero();

  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  int nd1 = d->get_axis(1).n;
  int nd2 = d->get_axis(2).n;
  int nd3 = d->get_axis(3).n;

  if(nd1 != _nd[0] || nd2 != _nd[1] || nd3 != _nd[2]) {
    seperr("lv_conv3df: data axes must be same in operator as in constructor.\n");
  }

  if(!d->check_match(_aux)) {
    seperr("lv_conv3df: input data and auxiliary should be same size.\n");
  }

  if(m->get_axis(1).n != _nlag || m->get_axis(2).n != _nf[0] || m->get_axis(3).n != _nf[1] || m->get_axis(4).n != _nf[2]) {
    seperr("lv_conv3df: input filter (model) is not same size as constructor.\n");
  }

  double dist1 = 0.0, dist2 = 0.0, dist3 = 0.0;
  /* Loop over each block */
  int fct = (_nf[0]-1)*(_nf[1]-1);
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    int k1 = ((ib)%(fct))%(_nf[0]-1);
    int k2 = ((ib)%(fct))/(_nf[0]-1);
    int k3 =  (ib)/fct;
    /* Get the filter coefficients at block corners */
    memcpy(_b000->vals,&m->vals[((k3+0)*_nf[1]*_nf[0] + (k2+0)*_nf[0] + k1+0)*_nlag],sizeof(float)*_nlag);
    memcpy(_b100->vals,&m->vals[((k3+0)*_nf[1]*_nf[0] + (k2+0)*_nf[0] + k1+1)*_nlag],sizeof(float)*_nlag);
    memcpy(_b110->vals,&m->vals[((k3+0)*_nf[1]*_nf[0] + (k2+1)*_nf[0] + k1+1)*_nlag],sizeof(float)*_nlag);
    memcpy(_b010->vals,&m->vals[((k3+0)*_nf[1]*_nf[0] + (k2+1)*_nf[0] + k1+0)*_nlag],sizeof(float)*_nlag);
    memcpy(_d001->vals,&m->vals[((k3+1)*_nf[1]*_nf[0] + (k2+0)*_nf[0] + k1+0)*_nlag],sizeof(float)*_nlag);
    memcpy(_d101->vals,&m->vals[((k3+1)*_nf[1]*_nf[0] + (k2+0)*_nf[0] + k1+1)*_nlag],sizeof(float)*_nlag);
    memcpy(_d111->vals,&m->vals[((k3+1)*_nf[1]*_nf[0] + (k2+1)*_nf[0] + k1+1)*_nlag],sizeof(float)*_nlag);
    memcpy(_d011->vals,&m->vals[((k3+1)*_nf[1]*_nf[0] + (k2+1)*_nf[0] + k1+0)*_nlag],sizeof(float)*_nlag);
    /* Compute axis 3 interpolating factors to be added */
    _d001->scale_add(1.0,_b000,-1.0); _d101->scale_add(1.0,_b100,-1.0);
    _d111->scale_add(1.0,_b110,-1.0); _d011->scale_add(1.0,_b010,-1.0);
    dist3 = 1.0/(_blocks[ib].e3 - _blocks[ib].b3 + 1.0);
    dist2 = 1.0/(_blocks[ib].e2 - _blocks[ib].b2 + 1.0);
    dist1 = 1.0/(_blocks[ib].e1 - _blocks[ib].b1 + 1.0);
    _d001->scale(dist3); _d101->scale(dist3);
    _d111->scale(dist3); _d011->scale(dist3);
    /* Loop over each point in the block */
    for(int id3 = _blocks[ib].b3; id3 <= _blocks[ib].e3; ++id3) {
      memcpy(_b00->vals,_b000->vals,sizeof(float)*_nlag);
      memcpy(_b10->vals,_b100->vals,sizeof(float)*_nlag);
      memcpy(_d01->vals,_b010->vals,sizeof(float)*_nlag);
      memcpy(_d00->vals,_b110->vals,sizeof(float)*_nlag);
      /* Compute axis 2 interpolating factors to be added */
      _d01->scale_add(1.0,_b000,-1.0);
      _d00->scale_add(1.0,_b100,-1.0);
      _d00->scale(dist2); _d01->scale(dist2);
      for(int id2 = _blocks[ib].b2; id2 <= _blocks[ib].e2; ++id2) {
        memcpy(_bin->vals,_b00->vals,sizeof(float)*_nlag);
        memcpy(_d0->vals,_b10->vals,sizeof(float)*_nlag);
        /* Compute axis 1 interpolating factors to be added */
        _d0->scale_add(1.0,_b00,-1.0);
        _d0->scale(dist1);
        for(int id1 = _blocks[ib].b1; id1 <= _blocks[ib].e1; ++id1) {
          for(int il = 0; il < _nlag; ++il) {
            d->vals[(id3+_lag[il].third)*nd1*nd2 + (id2+_lag[il].second)*nd1 + id1+_lag[il].first] +=
                _bin->vals[il]*_aux->vals[id3*nd2*nd1 + id2*nd1 + id1];
          }
          _bin->scale_add(1.0,_d0,1.0);
        }
        _b00->scale_add(1.0,_d01,1.0);
        _b10->scale_add(1.0,_d00,1.0);
      }
      _b000->scale_add(1.0,_d001,1.0);
      _b100->scale_add(1.0,_d101,1.0);
      _b110->scale_add(1.0,_d111,1.0);
      _b010->scale_add(1.0,_d011,1.0);
    }
  }

  return true;
}

void lv_conv3df::set_data(hypercube_float *data) {
  /* Create a filter with only zero-lag coefficients set */
  std::vector<axis> faxes;
  faxes.push_back(axis(_nlag ,0.0,1.0));
  faxes.push_back(axis(_nf[0],0.0,1.0));
  faxes.push_back(axis(_nf[1],0.0,1.0));
  faxes.push_back(axis(_nf[2],0.0,1.0));
  hypercube_float *tflt = new hypercube_float(faxes,true);
  tflt->zero();
  set_zerolag(tflt);
  /* Convolve auxiliary with this temporary filter */
  forward(false,tflt,data);
  delete tflt;
}

bool lv_conv3df::adjoint(bool add, my_vector *mod, my_vector *dat) {
  if(!add && mod) mod->zero();
  hypercube_float *m = (hypercube_float*)mod;
  hypercube_float *d = (hypercube_float*)dat;

  int nd1 = d->get_axis(1).n;
  int nd2 = d->get_axis(2).n;
  int nd3 = d->get_axis(3).n;

  if(nd1 != _nd[0] || nd2 != _nd[1] || nd3 != _nd[2]) {
    seperr("lv_conv3df: data axes must be same in operator as in constructor.\n");
  }

  if(!d->check_match(_aux)) {
    seperr("lv_conv3df: input data and auxiliary should be same size.\n");
  }

  if(m->get_axis(1).n != _nlag || m->get_axis(2).n != _nf[0] || m->get_axis(3).n != _nf[1] || m->get_axis(4).n != _nf[2]) {
    seperr("lv_conv3df: input filter (model) is not same size as constructor.\n");
  }

  double dist1 = 0.0, dist2 = 0.0, dist3 = 0.0;
  /* Loop over each block */
  int fct = (_nf[0]-1)*(_nf[1]-1);
  for(int ib = 0; ib < (int)_blocks.size(); ++ib) {
    int k1 = ((ib)%(fct))%(_nf[0]-1);
    int k2 = ((ib)%(fct))/(_nf[0]-1);
    int k3 =  (ib)/fct;
    _b000->zero(); _b100->zero(); _b110->zero(); _b010->zero();
    _d001->zero(); _d101->zero(); _d111->zero(); _d011->zero();
    /* Loop over each point in the block */
    for(int id3 = _blocks[ib].e3; id3 >= _blocks[ib].b3; --id3) {
      _b00->zero(); _b10->zero(); _d01->zero(); _d00->zero();
      for(int id2 = _blocks[ib].e2; id2 >=_blocks[ib].b2; --id2) {
        _d0->zero(); _bin->zero();
        for(int id1 = _blocks[ib].e1; id1 >= _blocks[ib].b1; --id1) {
          for(int il = 0; il < _nlag; ++il) {
            double x = d->vals[(id3+_lag[il].third)*nd1*nd2 + (id2+_lag[il].second)*nd1 + id1+_lag[il].first]*_aux->vals[id3*nd2*nd1 + id2*nd1 + id1];
            _d0->vals[il] += _bin->vals[il];
            _bin->vals[il] += x;
          }
        }
        dist1 = 1.0/(_blocks[ib].e1 - _blocks[ib].b1 + 1.0);
        dist2 = 1.0/(_blocks[ib].e2 - _blocks[ib].b2 + 1.0);
        dist3 = 1.0/(_blocks[ib].e3 - _blocks[ib].b3 + 1.0);
        _d01->scale_add(1.0,_b00,1.0);  //d21 += b21
        _d00->scale_add(1.0,_b10,1.0);  //d22 += b22
        _d0->scale(dist1);
        _bin->scale_add(1.0,_d0,-1.0);
        _b00->scale_add(1.0,_bin,1.0); //b21 += b1 - d1*dist1
        _b10->scale_add(1.0,_d0, 1.0); //b22 += d1*dist1
      }
      _d001->scale_add(1.0,_b000,1.0); //d31 += b31
      _d101->scale_add(1.0,_b100,1.0); //d32 += b32
      _d111->scale_add(1.0,_b110,1.0); //d33 += b33
      _d011->scale_add(1.0,_b010,1.0); //d34 += b34
      _d01->scale(dist2); _d00->scale(dist2);
      _b00->scale_add(1.0,_d01,-1.0);
      _b10->scale_add(1.0,_d00,-1.0);
      _b000->scale_add(1.0,_b00,1.0); //b31 += b21 - d21*dist2
      _b100->scale_add(1.0,_b10,1.0); //b32 += b22 - d22*dist2
      _b110->scale_add(1.0,_d00,1.0); //b33 += d22*dist2
      _b010->scale_add(1.0,_d01,1.0); //b34 += d21*dist2
    }
    for(int il = 0; il < _nlag; ++il) {
      m->vals[((k3+0)*_nf[1]*_nf[0] + (k2+0)*_nf[0] + k1+0)*_nlag + il] += -_d001->vals[il]*dist3 + _b000->vals[il];
      m->vals[((k3+0)*_nf[1]*_nf[0] + (k2+0)*_nf[0] + k1+1)*_nlag + il] += -_d101->vals[il]*dist3 + _b100->vals[il];
      m->vals[((k3+0)*_nf[1]*_nf[0] + (k2+1)*_nf[0] + k1+1)*_nlag + il] += -_d111->vals[il]*dist3 + _b110->vals[il];
      m->vals[((k3+0)*_nf[1]*_nf[0] + (k2+1)*_nf[0] + k1+0)*_nlag + il] += -_d011->vals[il]*dist3 + _b010->vals[il];
      m->vals[((k3+1)*_nf[1]*_nf[0] + (k2+0)*_nf[0] + k1+0)*_nlag + il] +=  _d001->vals[il]*dist3;
      m->vals[((k3+1)*_nf[1]*_nf[0] + (k2+0)*_nf[0] + k1+1)*_nlag + il] +=  _d101->vals[il]*dist3;
      m->vals[((k3+1)*_nf[1]*_nf[0] + (k2+1)*_nf[0] + k1+1)*_nlag + il] +=  _d111->vals[il]*dist3;
      m->vals[((k3+1)*_nf[1]*_nf[0] + (k2+1)*_nf[0] + k1+0)*_nlag + il] +=  _d011->vals[il]*dist3;
    }
  }
  return true;
}
