/**
 * A static class for plotting data from within a C++ program
 * using the gnuplot-iostream library written
 * by Dan Stahlke et al (a Physicist at Carnegie Mellon)
 * https://github.com/dstahlke/gnuplot-iostream/wiki
 * @author: Joseph Jennings, Stanford University
 * @version: 2017.01.06
 */
#ifndef SIMPLEPLOT_H_
#define SIMPLEPLOT_H_

#include "hypercube_float.h"
#include "hypercube_complex.h"
#include "axis.h"
#include <vector>

namespace simpleplot{

  /**
   * A generic N-D plotting function
   * @param data the input dataset
   */
  void plot(hypercube_float *data);

  /**
   * Plot a 1D dataset
   * @param data the input dataset
   */
  void plot1d(hypercube_float *data);

  /**
   * Overlay two 1D plots
   * @param data1 the first input data
   * @param data2 the second input data
   * @param flag whether to clip to signal one or two
   */
  void plot1d2(hypercube_float *data1, hypercube_float *data2, bool clip1);


  /**
   * Plot a 1D dataset
   * @param data the input dataset
   */
  void plot1d(my_vector *data);

  /**
   * Plot a 1D dataset as impulses
   * @param data the input dataset
   */
  void plot1dimp(hypercube_float *data);

  /**
   * Overlay two impulse plots
   * @param data1 the first input data
   * @param data2 the second input data
   * @param flag whether to clip to signal one or two
   */
  void plot1dimp2(hypercube_float *data1, hypercube_float *data2, bool clip1);

  /**
   * Overlay three impulse plots
   * @param data1 the first input data
   * @param data2 the second input data
   * @param data3 the third input data
   * @param wichclip number whether to clip to one,two or three
   */
  void plot1dimp3(hypercube_float *data1, hypercube_float *data2, hypercube_float *data3, int wichclp);

  /** Plot the amplitude spectrum of a 1D dataset
   * @param data the input dataset
   */
  void plot1dspec(hypercube_float *data);

  /**
   * Plot a 1D dataset
   * @param data an input vector of floats
   * @param myaxis the axis of the input data
   */
  void plot1d(std::vector<float> data, axis myaxis);

  /**
   * Plot a 1D dataset
   * @param data an input vector of floats
   * @param myaxis the axis of the input data
   */
  void plot1dimp(std::vector<float> data, axis myaxis);

  /**
   * Plot a 2D dataset
   * @param data the input data
   */
  void plot2d(hypercube_float *data);

  /**
   * Plot a 2D dataset
   * @param data the input data
   */
  void plot2d(my_vector *data);

  /**
   * Plot a 2D dataset
   * @param data the input dataset
   */
  void plot2d(norm_vector *data);

  /**
   * Plot a 2D dataset
   * @param data the input dataset
   * @param axes the axes of the dataset
   */
  void plot2d(std::vector<float> data, std::vector<axis> axes);

  void animate(hypercube_float *data);
  void animate1d(hypercube_float *data);
  void animate2d(hypercube_float *data);
  std::string get_range(hypercube_float *data);
  std::string get_range(std::vector<float> data, axis myaxis);
  std::string get_range2d();
  std::string get_range2d(hypercube_float *data);
  std::string get_range2d(std::string cscale);
  void plot_slice1d1(hypercube_float *data);
  void plot_slice1d2(hypercube_float *data);
  void plot_slice1d3(hypercube_float *data);
  void plot_slice2d1(hypercube_float *data);
  void plot_slice2d2(hypercube_float *data);
  void plot_slice2d3(hypercube_float *data, int frame);
}

#endif /* SIMPLEPLOT_H_ */
