extern "C" {
#include "seplib.h"
}
#include <complex>
#include <random>
#include <fftw3.h>
#include "wavelet.h"
#include "simpleplot.h"

wavelet::wavelet(axis st) {
  _ntsrc = st.n;
  _dtsrc = st.d;
  _ftsrc = st.o;
}

wavelet::wavelet(axis st, float dly) {
  _ntsrc = st.n;
  _dtsrc = st.d;
  _ftsrc = st.o;
  _dly   = dly;
}

void wavelet::ricker(float f, hypercube_float *src) {
  for (int it = 0;  it < _ntsrc; ++it) {
    float t = it * _dtsrc - _dly;
    float pift = M_PI * M_PI * (f * f) * (t * t);
    src->vals[it] = (1 - 2 * pift) * exp(-pift);
  }
}

void wavelet::monosin(float f, float amp, float ph, hypercube_float *src) {
  float phrad = ph*DEG2RAD;
  for(int it = 0; it < _ntsrc; ++it) {
    float t = it * _dtsrc;
    src->vals[it] = amp*cos(f * 2*M_PI* t - phrad);
  }
}

void wavelet::sumsin(std::vector<float> fs, std::vector<float> amps, std::vector<float>phs, hypercube_float* src) {
  int nsig = fs.size();
  src->zero(); std::vector<float> phrads;
  /* Convert to radians */
  for(int is = 0; is < nsig; ++is) {phrads.push_back(phs[is]*DEG2RAD);}
  if(nsig != (int)amps.size() && nsig != (int)phs.size()) seperr("Num freqs, amps and phzs must be the same\n.");
  for(int it = 0; it < _ntsrc; ++it) {
    for(int is = 0; is < nsig; ++is) {
      float t = it * _dtsrc;
      src->vals[it] += amps[is]*cos(fs[is]* 2*M_PI * t - phrads[is]);
    }
  }
}

void wavelet::bprandom(float f1, float f2, float f3, float f4, hypercube_float *src) {
  std::random_device rd; std::mt19937 rng(rd());
  std::uniform_real_distribution<float> uni(-0.5,0.5);
  for(int it = 0; it < _ntsrc; ++it) {
    auto val = uni(rng);
    src->vals[it] = val;
  }
  bandpass(f1,f2,f3,f4,src);
}

void wavelet::bprandom(std::vector<float> freqs, hypercube_float *src) {
  bprandom(freqs[0],freqs[1],freqs[2],freqs[3],src);
}

void wavelet::bandpass(float f1, float f2, float f3, float f4, hypercube_float* src) {
  int nw   = _ntsrc/2;
  float dw = 1/(_ntsrc*_dtsrc);
  std::complex<double> *in   = new std::complex<double>[_ntsrc];
  std::complex<double> *in2  = new std::complex<double>[_ntsrc];
  std::complex<double> *out  = new std::complex<double>[_ntsrc];

  for(int it = 0; it < _ntsrc; ++it) {
    in[it] = std::complex<double>(static_cast<double>(src->vals[it]),0.0);
  }
  /* Compute the FFT */
  fftw_plan p = fftw_plan_dft_1d(_ntsrc,reinterpret_cast<fftw_complex*>(in),
      reinterpret_cast<fftw_complex*>(out),FFTW_FORWARD,FFTW_ESTIMATE);
  fftw_execute(p);

  /* Apply band pass */
  std::complex<double> coef; double coefr; double coefi = 0.0;
  for(int iw = 0; iw < nw; ++iw) {
    float freq = dw*iw;
    if(freq < f1) {
      coefr             = 0.0;
      coef              = std::complex<double>(coefr,coefi);
      out[iw]          *= coef;
      out[_ntsrc-1-iw] *= coef;
    } else if(freq >= f1 && freq < f2) {
      coefr             = cos(M_PI/2.0 * (f2-freq)/(f2-f1))*cos(M_PI/2.0 * (f2-freq)/(f2-f1));
      coef              = std::complex<double>(coefr,coefi);
      out[iw]          *= coef;
      out[_ntsrc-1-iw] *= coef;
    } else if(freq >= f2 && freq < f3) {
      coefr             = 1.0;
      coef              = std::complex<double>(coefr,coefi);
      out[iw]          *= coef;
      out[_ntsrc-1-iw] *= coef;
    } else if(freq >= f3 && freq < f4){
      coefr             = cos(M_PI/2.0 * (freq-f3)/(f4-f3))*cos(M_PI/2.0 * (freq-f3)/(f4-f3));
      coef              = std::complex<double>(coefr,coefi);
      out[iw]          *= coef;
      out[_ntsrc-1-iw] *= coef;
    } else {
      coefr             = 0.0;
      coef              = std::complex<double>(coefr,coefi);
      out[iw]          *= coef;
      out[_ntsrc-1-iw] *= coef;
    }
  }

  /* Inverse FFT */
   fftw_plan q = fftw_plan_dft_1d(_ntsrc,reinterpret_cast<fftw_complex*>(out),
       reinterpret_cast<fftw_complex*>(in2),
       FFTW_BACKWARD,FFTW_ESTIMATE);

   fftw_execute(q);

   /* Copy filtered version to source */
   for(int it = 0; it < _ntsrc; ++it) src->vals[it] = in2[it].real();
   src->scale(static_cast<float>(1.0/_ntsrc));
}

void wavelet::bandpass(std::vector<float> freqs, hypercube_float* src) {
  if(freqs.size() != 4)
    seperr("Bandpass requires 4 frequencies. You entered %d.\n",freqs.size());
  bandpass(freqs[0],freqs[1],freqs[2],freqs[3],src);
}

void wavelet::gaussian(float a, float c, hypercube_float *src) {
  float scale = 1/(2*c*c);
  for (int it = 0;  it < _ntsrc; ++it) {
    float t = it * _dtsrc - _dly;
    src->vals[it] = a * exp(-t*t*scale);
  }
}

void wavelet::expdecay(float a, float b, hypercube_float *src) {
  float swtch = 0.0;
  for(int it = 0; it < _ntsrc; ++it) {
    float t = it * _dtsrc;
    if(t < _dly) {
      swtch = 0.0;
    } else {
      swtch = 1.0;
      t -= _dly;
    }
    src->vals[it] = swtch*a*exp(-b*t);
  }
}
