# 1D stationary adaptive filter
import sys, argparse, ConfigParser
import seppy
import numpy as np

class GradMSE:
  """ Class for minimizing MSE """

  def __init__(self,nw,m,x,d,dshift):
    """ Constructs auto and cross correlation matrices for optimization """
    assert(m != 0), "Cannot perform steepest descent with m=0.\n"
    self.m = m
    self.nw = nw
    self.dshift = dshift
    # Construct P
    self.P = np.zeros(nw)
    for k in range(nw,m+nw):
      xk = np.flip(x[k-nw:k],0)
      self.P += d[k - nw + self.dshift]*xk
    self.P /= self.m

    # Construct R
    self.R = np.zeros([nw,nw])
    for k in range(nw,m+nw):
      xk = np.flip(x[k-nw:k],0)
      self.R += np.outer(xk,xk)
    self.R /= self.m

    # Construct expectation over data
    self.edk2 = 0
    for k in range(nw,m+nw):
      self.edk2 += d[k-nw+self.dshift]*d[k-nw+self.dshift]
    self.edk2 /= self.m

  def getR(self):
    """ Returns the autocorrelation matrix """
    return self.R

  def getP(self):
    """ Returns the cross-correlation vector """
    return self.P

  def getedk2(self):
    """ Returns the MS of the data """
    return self.edk2

  def calcMSE(self,w):
    """ Calculates the mean squared error for a particular w """
    wtR = np.dot(w,self.R)
    return self.edk2 + np.dot(wtR,w) - 2*np.dot(self.P,w)

  def calcGrad(self,w):
    """ Calculates the gradient of the mean squared error for w """
    return 2*np.dot(self.R,w) - 2*self.P

  def calcError(self,x,d,w):
    """ Calculates the error for the current w """
    err = np.zeros(len(d))
    #print x
    #print d
    for k in range(self.nw,self.m + self.nw):
      xk = np.flip(x[k-self.nw:k],0)
      #print xk, d[k-self.nw + self.dshift], np.dot(xk,w)
      err[k-nw] = d[k - self.nw + self.dshift] - np.dot(xk,w)

    return err

  def solveSD(self,w0,mu,niter):
    """ Performs steepest descent and returns w* and MSEmin """
    witer = np.zeros(len(w0))
    witer[:] = w0
    for kiter in range(niter):
      xi0 = self.calcMSE(witer)
      witer = witer +  mu * -self.calcGrad(witer)
      xi1 = self.calcMSE(witer)
      if(xi1 >= xi0): 
         print "MSE not reduced, terminating solver..."
         break
      print "iter=%d MSE=%f"%(kiter,xi1)

    return witer,xi1

  def solveNewt(self,w0,mu,niter):
    """ Performs Newton's method and returns w* and MSEmin """
    witer = np.zeros(len(w0))
    witer[:] = w0
    Rinv = np.linalg.inv(self.R)
    for kiter in range(niter):
      xi0 = self.calcMSE(witer)
      witer = witer + mu * np.dot(Rinv,-self.calcGrad(witer))
      xi1 = self.calcMSE(witer)
      if(xi1 >= xi0):
          print "MSE not reduced, terminating solver..."
          break;
      print "iter=%d MSE=%f"%(kiter,xi1)

    return witer,xi1

if __name__ == '__main__':
  # Parse the config file
  conf_parser = argparse.ArgumentParser(add_help=False)
  conf_parser.add_argument("-c", "--conf_file",
                           help="Specify config file", metavar="FILE")
  args, remaining_argv = conf_parser.parse_known_args()
  defaults = { 
      "alg": "newt",
      "w0": None,
      "mu": 0.5,
      "niter": 100,
      }   
  if args.conf_file:
    config = ConfigParser.SafeConfigParser()
    config.read([args.conf_file])
    defaults = dict(config.items("Defaults"))
    
  # Parse the other arguments
  # Don't surpress add_help here so it will handle -h
  parser = argparse.ArgumentParser(parents=[conf_parser],description=__doc__,
      formatter_class=argparse.ArgumentDefaultsHelpFormatter)
  
  parser.set_defaults(**defaults)
  # Input files
  parser.add_argument("in=",help="Input signal")
  parser.add_argument("din=",help="Desired signal")
  parser.add_argument("out=",help="Output filter")
  # Required arguments
  requiredNamed = parser.add_argument_group('required parameters')
  requiredNamed.add_argument("-nw",help="Length of filter",required=True,type=int)
  # Optional arguments
  parser.add_argument("-mu",help="Adaptation rate",type=float)
  parser.add_argument("-niter",help="Number of iterations",type=int)
  parser.add_argument("-w0",help="Initial w guess")
  parser.add_argument("-err",help="Output error signal")
  parser.add_argument("-alg",help="Algorithm for estimating filter")
  parser.add_argument("-verb",help="verbosity flag")
  args = parser.parse_args(remaining_argv)

  # Read in input arguments
  nw = args.nw
  alg = args.alg
  verb = args.verb
  mu = args.mu
  niter = args.niter
  w0 = np.zeros(nw)
  wopt = np.zeros(nw)
  msemin = 0.0
  errflg = args.err
  if(args.w0 != None):
    for iw in args.w0.split(','): w0[iw] = float(iw)
  
  # Set up SEP
  sep = seppy.sep(sys.argv)
  saxes,sig = sep.read_file("in")
  daxes,des = sep.read_file("din")
  
  nd = saxes.n[0]
  err = None
  assert(nd == daxes.n[0]), "Input and desired signal must be same length.\n"
  
  if(np.sum(sig == des) == nd):
    if(verb == "y"): print "Performing autogression (PEF estimation)"
    # Create all x examples (1 to m)
    m = nd - nw
    #X = np.zeros([m,nw])
    #for i in range(nw,m+nw):
    #  X[i-nw] = np.flip(sig[i-nw:i],0)
  
    #print sig
    #print des
    #print des.shape
    #for i in range(nw,m+nw):
    #  print X[i-nw],des[i]

    #print X
    if(alg == "newt"):
      gd = GradMSE(nw,m,sig,des,nw)
      wopt, msemin = gd.solveNewt(w0,mu,niter)
      if(errflg != None): err = gd.calcError(sig,des,wopt)

    if(alg == "sd"):
      gd = GradMSE(nw,m,sig,des,nw)
      wopt, msemin = gd.solveSD(w0,mu,niter)
      if(errflg != None): err = gd.calcError(sig,des,wopt)
  
  else:
    if(verb =="y"): print "Performing regression (shaping filter estimation)"
    # m examples
    m = nd - nw + 1

    if(alg == "newt"):
      print nw, m
      gd = GradMSE(nw,m,sig,des,1)
      print gd.getR()
      wopt, msemin = gd.solveNewt(w0,mu,niter)
      if(errflg != None): err = gd.calcError(sig,des,wopt)

    if(alg == "sd"):
      gd = GradMSE(nw,m,sig,des,1)
      wopt, msemin = gd.solveSD(w0,mu,niter)
      if(errflg != None): err = gd.calcError(sig,des,wopt)
  
  # Write out optimal filters
  waxes = seppy.axes([nw],[0.0],[1.0])
  sep.write_file("out",waxes,wopt)

  # Write out error 
  if(errflg != None):
    sep.write_file("-err",saxes,err)

