/**
 * Generates a 1D wavelet (point source)
 * for wave propagation
 * @author: Joseph Jennings
 * @version: 2016.12.7
 */

#include <algorithm>
#include <sep_params.h>
#include <axis.h>
#include "wavelet.h"
#include "simpleplot.h"

#define MY_SEP_DOC \
    sep_add_doc_line("NAME");\
    sep_add_doc_line("    GenWavelet - Create a 1D band-limited wavelet (point source) for finite-difference wave propagation");\
    sep_add_doc_line(""); \
    sep_add_doc_line("SYNOPSIS");\
    sep_add_doc_line("    GenWavlelet nt= dt= dly= freqs=f1,f2,... mode=[rick],rand,mono,sums,gauss > out.H"); \
    sep_add_doc_line(""); \
    sep_add_doc_line("INPUT PARAMETERS");\
    sep_add_doc_line("    nt  - integer"); \
    sep_add_doc_line("        number of time samples on source wavelet."); \
    sep_add_doc_line("    dt  - float"); \
    sep_add_doc_line("        wavelet time sampling interval."); \
    sep_add_doc_line("    dly  - float"); \
    sep_add_doc_line("        amount of how much to shift the wavelet in time."); \
    sep_add_doc_line("    mode - char*"); \
    sep_add_doc_line("        mode for creating wavelet."); \
    sep_add_doc_line("        rick is for ricker wavelet [default]."); \
    sep_add_doc_line("        mono is for monochromatic sinusoid."); \
    sep_add_doc_line("        sums is for multiple frequency sinusoid."); \
    sep_add_doc_line("        gauss is a gaussian wavelet."); \
    sep_add_doc_line("        rnd band-limited random wavelet."); \
    sep_add_doc_line("    freqs  - floats"); \
    sep_add_doc_line("        for rick and mono mode, use one frequency."); \
    sep_add_doc_line("        for sums use up to 10 frequencies."); \
    sep_add_doc_line("        for rnd use four frequencies to define the band."); \
    sep_add_doc_line(""); \

#ifndef SOURCE
#define SOURCE "/home/joe/phd/projects/lwi_pf/Src/test/GenWavelet.cpp"
#endif

int main(int argc, char **argv) {

  /* Initialize IO */
  MY_SEP_DOC;
  sep_params pars = sep_params(argc, argv, SOURCE);

  /* Get parameters */
  int maxsig = 10;
  int nt = pars.get_int("nt", 0); if(!nt) seperr("Could not get n1\n");
  float dt   = pars.get_float("dt"  , 0.004);
  float ot   = pars.get_float("ot"  , 0.0);
  float dly  = pars.get_float("dly" , 1.0);
  std::vector<float> frqs = pars.get_vfloats("freqs",maxsig);
  std::vector<float> amps  = pars.get_vfloats("amps",maxsig);
  std::vector<float> phzs  = pars.get_vfloats("phzs",maxsig);
  std::string mode = pars.get_string("mode","rick");
  /* Defaults for one signal */
  if(frqs.empty()) frqs.push_back(30.0);
  if(amps.empty()) amps.push_back(1.0);
  if(phzs.empty()) phzs.push_back(0.0);
  /* Defaults for > 1 signal */
  int nsig = std::max({frqs.size(),amps.size(),phzs.size()});
  if((int)frqs.size() < nsig) {
    for(int i = frqs.size(); i < nsig; ++i) frqs.push_back(30.0);
  }
  if((int)amps.size() < nsig) {
    for(int i = amps.size(); i < nsig; ++i) amps.push_back(1.0);
  }
  if((int)phzs.size() < nsig) {
    for(int i = phzs.size(); i < nsig; ++i) phzs.push_back(0.0);
  }

  /* Source time axis */
  axis st = axis(nt,ot,dt);
  std::vector<axis> saxis(1); saxis[0] = st;

  hypercube_float *src = new hypercube_float(saxis, true);

  wavelet wav = wavelet(st, dly);

  //TODO: Add min or max phase
  /* Compute wavelet based on mode */
  if(!mode.compare("rick")) {
    wav.ricker(frqs[0], src);
  } else if(!mode.compare("rand")) {
    wav.bprandom(frqs, src);
  } else if(!mode.compare("mono")) {
    wav.monosin(frqs[0],amps[0],phzs[0],src);
  } else if(!mode.compare("sums")) {
    wav.sumsin(frqs,amps,phzs,src);
  } else if(!mode.compare("gauss")) {
    wav.gaussian(amps[0],frqs[0],src);
  } else if(!mode.compare("exp")) {
    wav.expdecay(amps[0],frqs[0],src);
  } else {
    seperr("Mode: %s not recognized. Please try another mode.\n",mode.c_str());
  }

  sregf out = sregf(saxis);

  out.write_description("out");
  out.write_all("out", src);

  delete src;
}
