/**
 * Tests the lv_conv1dfmc forward operator
 * @author: Joseph Jennings
 * @version: 2018.02.07
 */

#include "sregf.h"
#include "hypercube_float.h"
#include "../main/config_parser.h"
#include "../main/error_handle.h"
#include "../main/lv_conv1dfmc.h"

int main(int argc, char **argv) {

  /* Get command line parameters */
  ConfigParser cfg = ConfigParser(argc, argv, true);
  int j = cfg.get_int("cmd", "j", 10);
  int nt = cfg.get_int("cmd", "nd", 1000);
  int nc = cfg.get_int("cmd", "nc", 2);
  int spikeloc = cfg.get_int("cmd","sploc",499);
  if(spikeloc >= nt) { throw_invalid_arg("Spike location must be less than data size.\n"); }
  bool verb = cfg.get_int("cmd","verb",true);
  //TODO: Really should catch the exceptions here
  std::vector<int> nlag;
  std::vector<int> lag1; std::vector<int> lag2;
  try{
    nlag = cfg.get_ints("cmd","nlag");
    if(nlag[1] > nc) {
      fprintf(stderr,"Cannot have more lags than components.\n");
      return 1;
    }
  } catch(const InvalidArg &ex) {
    try {
    lag1 = cfg.get_ints("cmd","lag1");
    lag2 = cfg.get_ints("cmd","lag2");
    if(lag1.size() != lag2.size()) { fprintf(stderr,"Must have equal number of lag1 and lag2.\n"); return 1; }
    } catch(const InvalidArg &ex) {
      fprintf(stderr,"Must pass either nlags or lag coordinates.\n");
      return 1;
    }
  }

  std::vector<axis> spaxes;
  spaxes.push_back(axis(nt,0.0,1.0));
  spaxes.push_back(axis(nc,0.0,1.0));

  hypercube_float *spike = new hypercube_float(spaxes,true);
  spike->vals[spikeloc] = 1.0;
  hypercube_float *imp = new hypercube_float(spaxes,true);

  /* Build the convolution operator */
  lv_conv1dfmc *lv1mc = nullptr;
  if(!nlag.empty()) {
    lv1mc = new lv_conv1dfmc(nt,j,nlag,spike);
  } else if(!lag1.empty() && !lag2.empty()) {
    if(lag1.size() != lag2.size()) {
      seperr("Lag 1 and lag 2 are not the same size. Each lag point "
          "requires one coordinate (axis 1 and axis 2).\n");
    }
    std::vector<std::pair<int,int>> lags;
    for(int il = 0; il < (int)lag1.size(); ++il) {
      lags.push_back(std::pair<int,int>(lag1[il],lag2[il]));
    }
    lv1mc = new lv_conv1dfmc(nt,j,lags,spike);
  }

  if(verb) {
    lv1mc->print_blockinfo();
    fprintf(stderr,"\n");
    lv1mc->print_laginfo();
  }

  std::vector<int> nf = lv1mc->get_numfilt();
  std::vector<axis> faxes;
  faxes.push_back(axis(nf[0],0.0,1.0));
  faxes.push_back(axis(nf[1],0.0,1.0));

  hypercube_float *flt  = new hypercube_float(faxes,true);
  flt->random();
  lv1mc->set_zerolag(flt);
  lv1mc->forward(false,flt,imp);

  /* Write out impulse response */
  sregf impout = sregf(imp->get_axes());
  impout.write_description("out");
  impout.write_all("out", imp);

}
