/**
 * Estimates linearly-varying,
 * non-stationary PEFs on a 1D signal
 * via conjugate gradient iterations
 * @author: Joseph Jennings
 * @version: 2017.12.21
 */

#include "sep_params.h"
#include "combo_oper.h"
#include "../main/lv_conv1df.h"
#include "../main/lv_conv1dmask.h"
#include "../main/identity.h"
#include "../main/conj_grad.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);
  int niter = pars.get_int("niter",100);
  int verb = pars.get_int("verb",0);
  float eps = pars.get_float("eps",0.01);
  std::string rmov = pars.get_string("rmov","");
  std::string mmov = pars.get_string("mmov","");
  std::string gmov = pars.get_string("gmov","");
  int maxlag = 50;
  std::vector<int> lags = pars.get_vints("lag",maxlag);
  int nlag = pars.get_int("nlag",0);
  int j    = pars.get_int("j",0);
  if(lags.empty() && nlag == 0) {
    seperr("Please supply either lag coordinates or the number of lags\n");
  }
  if(j == 0) {
    seperr("Please supply filter coarse sampling (j's).\n");
  }

  sregf sigin = sregf("in");
  hypercube_float *sig = new hypercube_float(sigin.get_axes(),true);
  hypercube_float *dat = new hypercube_float(sigin.get_axes(),true);
  sigin.read_all("in",sig);

  int nd  = sig->get_axis(1).n;

  /* Build the convolution operator */
  lv_conv1df *lv1 = nullptr;
  if(nlag != 0) {
    lv1 = new lv_conv1df(nd,j,nlag,sig);
  } else if(!lags.empty()) {
    lv1 = new lv_conv1df(nd,j,lags,sig);
  }
  if(verb) {
    lv1->print_blockinfo();
    fprintf(stderr,"\n");
    lv1->print_laginfo();
  }
  /* Create data and model vectors */
  lv1->set_data(dat); dat->scale(-1.0);
  std::vector<int> nf = lv1->get_numfilt();
  std::vector<axis> faxes;
  faxes.push_back(axis(nf[0],0.0,1.0));
  faxes.push_back(axis(nf[1],0.0,1.0));
  hypercube_float *flt  = new hypercube_float(faxes,true);
  flt->zero();
  lv1->set_zerolag(flt);
  hypercube_float *zro = new hypercube_float(faxes,true);
  zro->zero();

  /* Create masked operator */
  lv_conv1dmask *k = new lv_conv1dmask();
  lv1->set_domain(flt); lv1->set_range(dat);
  k->set_domain(flt); k->set_range(flt);
  chain_op *dk = new chain_op(k,lv1);

  /* Create the identity operator */
  identity *iop = new identity();
  iop->set_domain(flt); iop->set_range(flt);

  /* Run the regularized inversion */
  conj_grad cg = conj_grad(niter);
  if(!rmov.empty()) {
    cg.set_resout("rmov");
  } if(!gmov.empty()) {
    cg.set_grdout("gmov");
  } if(!mmov.empty()) {
    cg.set_modout("mmov");
  }
  cg.run(dk,iop,dat,zro,flt,eps);
  lv1->set_zerolag(flt);
  sregf pefout = sregf(flt->get_axes());
  pefout.write_description("out");
  pefout.write_all("out",flt);

  /* Write out filter parameters */
  printf("j=%d\n",j);
  if(nlag != 0) {
    printf("nlag=%d",nlag);
  } else {
    for(int il = 0; il < (int)lags.size(); ++il) {
      printf("lag=%d,",lags[il]);
    }
  }

  /* Clean up */
  delete sig; delete dat; delete flt; delete zro;
}
