/**
 * Estimates linearly-varying,
 * non-stationary PEFs on a multcomponent
 * time-series via conjugate gradient iterations
 * @author: Joseph Jennings
 * @version: 2018.02.11
 */

#include "sep_params.h"
#include "combo_oper.h"
#include "../main/lv_conv1dfmc.h"
#include "../main/lv_conv1dmask.h"
#include "../main/identity.h"
#include "../main/conj_grad.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);
  int niter = pars.get_int("niter",100);
  int verb = pars.get_int("verb",0);
  float eps = pars.get_float("eps",0.01);
  std::string rmov = pars.get_string("rmov","");
  std::string mmov = pars.get_string("mmov","");
  std::string gmov = pars.get_string("gmov","");
  int maxlag = 50;
  std::vector<int> lag1 = pars.get_vints("lag1",maxlag);
  std::vector<int> lag2 = pars.get_vints("lag2",maxlag);
  std::vector<int> nlag = pars.get_vints("nlag",2);
  int j                 = pars.get_int("j",0);
  if(lag1.empty() && lag2.empty() && nlag.empty()) {
    seperr("Please supply either lag coordinates or the filter box dimensions\n");
  }
  if(j == 0) {
    seperr("Please supply filter coarse sampling (j).\n");
  }

  sregf sigin = sregf("in");
  hypercube_float *sig = new hypercube_float(sigin.get_axes(),true);
  hypercube_float *dat = new hypercube_float(sigin.get_axes(),true);
  sigin.read_all("in",sig);

  int nd = sigin.get_axis(1).n;

  /* Build the convolution operator */
  lv_conv1dfmc *lv1mc = nullptr;
  if(!nlag.empty()) {
    lv1mc = new lv_conv1dfmc(nd,j,nlag,sig);
  } else if(!lag1.empty() && !lag2.empty()) {
    if(lag1.size() != lag2.size()) {
      seperr("Lag 1 and lag 2 are not the same size. Each lag point "
          "requires one coordinate (axis 1 and axis 2).\n");
    }
    std::vector<std::pair<int,int>> lags;
    for(int il = 0; il < (int)lag1.size(); ++il) {
      lags.push_back(std::pair<int,int>(lag1[il],lag2[il]));
    }
    lv1mc = new lv_conv1dfmc(nd,j,lags,sig);
  }
  if(verb) {
    lv1mc->print_blockinfo();
    fprintf(stderr,"\n");
    lv1mc->print_laginfo();
  }
  /* Create data and model vectors */
  lv1mc->set_data(dat); dat->scale(-1.0);
  std::vector<int> nf = lv1mc->get_numfilt();
  std::vector<axis> faxes;
  faxes.push_back(axis(nf[0],0.0,1.0));
  faxes.push_back(axis(nf[1],0.0,1.0));
  hypercube_float *flt  = new hypercube_float(faxes,true);
  flt->zero();
  lv1mc->set_zerolag(flt);
  hypercube_float *zro = new hypercube_float(faxes,true);
  zro->zero();

  /* Create masked operator */
  lv_conv1dmask *k = new lv_conv1dmask();
  lv1mc->set_domain(flt); lv1mc->set_range(dat);
  k->set_domain(flt); k->set_range(flt);
  chain_op *dk = new chain_op(k,lv1mc);

  /* Create the identity operator */
  identity *iop = new identity();
  iop->set_domain(flt); iop->set_range(flt);

  /* Run the regularized inversion */
  conj_grad cg = conj_grad(niter);
  if(!rmov.empty()) {
    cg.set_resout("rmov");
  } if(!gmov.empty()) {
    cg.set_grdout("gmov");
  } if(!mmov.empty()) {
    cg.set_modout("mmov");
  }
  cg.run(dk,iop,dat,zro,flt,eps);
  lv1mc->set_zerolag(flt);
  sregf pefout = sregf(flt->get_axes());
  pefout.write_description("out");
  pefout.write_all("out",flt);

  /* Write out filter parameters */
  printf("j=%d\n",j);
  if(!nlag.empty()) {
    printf("nlag=%d,%d",nlag[0],nlag[1]);
  } else {
    for(int il = 0; il < (int)lag1.size(); ++il) {
      printf("lag1=%d,",lag1[il]);
    }
    printf("\n");
    for(int il = 0; il < (int)lag2.size(); ++il) {
      printf("lag2=%d,",lag2[il]);
    }
  }

  delete sig; delete dat; delete flt; delete zro;
}
