/**
 * Estimates linearly-varying,
 * non-stationary PEFs on a 3D dataset
 * via conjugate gradient iterations
 * @author: Joseph Jennings
 * @version: 2017.03.23
 */

#include "sep_params.h"
#include "combo_oper.h"
#include "../main/lv_conv3df.h"
#include "../main/lv_conv3dmask.h"
#include "../main/identity.h"
#include "../main/conj_grad.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc,argv);
  int niter = pars.get_int("niter",100);
  int vrb = pars.get_int("verb",0);
  float eps = pars.get_float("eps",0.01);
  int maxlag = 100; std::vector<int> empt;
  std::string rmov = pars.get_string("rmov","");
  std::string mmov = pars.get_string("mmov","");
  std::string gmov = pars.get_string("gmov","");
  std::vector<int> lag1 = pars.get_vints("lag1",maxlag);
  std::vector<int> lag2 = pars.get_vints("lag2",maxlag);
  std::vector<int> lag3 = pars.get_vints("lag3",maxlag);
  std::vector<int> nlag = pars.get_vints("nlag",3);
  std::vector<int> j    = pars.get_vints("j",3);
  if(lag1.empty() && lag2.empty() && lag3.empty()) {
    if(nlag.empty()) {
      seperr("Please supply either lag coordinates or the filter box dimensions\n");
    }
    if(nlag.size() != 3) {
      seperr("Please supply three nlag values or a list of lag coordinates\n");
    }
  }
  if(j.empty()) {
    seperr("Please supply filter coarse sampling (j's).\n");
  }
  if(j.size() != 3) {
    seperr("Please supply three j values\n");
  }

  sregf cubin = sregf("in");
  hypercube_float *cub = new hypercube_float(cubin.get_axes(),true);
  hypercube_float *dat = new hypercube_float(cubin.get_axes(),true);
  cubin.read_all("in",cub);

  std::vector<int> nd;
  nd.push_back(cub->get_axis(1).n);
  nd.push_back(cub->get_axis(2).n);
  nd.push_back(cub->get_axis(3).n);

  /* Build the convolution operator */
  lv_conv3df *lv3 = nullptr;
  if(!nlag.empty()) {
    lv3 = new lv_conv3df(nd,j,nlag,cub);
  }
  else if(!lag1.empty() && !lag2.empty() && !lag3.empty()) {
    if(lag1.size() != lag2.size() || lag3.size() != lag2.size()) {
      seperr("Lag 1, lag 2 and lag 3 are not the same size. Each lag point "
          "requires one coordinate (axis 1, axis 2 and axis 3).\n");
    }
    std::vector<triple> lags;
    for(int il = 0; il < (int)lag1.size(); ++il) {
      lags.push_back(triple(lag1[il],lag2[il],lag3[il]));
    }
    lv3 = new lv_conv3df(nd,j,lags,cub);
  }
  if(vrb) {
    lv3->print_blockinfo();
    fprintf(stderr,"\n");
    lv3->print_laginfo();
    std::vector<axis> taxes = lv3->get_filtaxes();
    fprintf(stderr,"nf1=%d nf2=%d nf3=%d\n",taxes[1].n,taxes[2].n,taxes[3].n);
    fprintf(stderr,"\n");
  }

  /* Create data and model vectors */
  lv3->set_data(dat); dat->scale(-1.0);
  hypercube_float *flt  = new hypercube_float(lv3->get_filtaxes(),true);
  flt->zero();
  lv3->set_zerolag(flt);
  hypercube_float *zro = new hypercube_float(lv3->get_filtaxes(),true);
  zro->zero();

  /* Create masked operator */
  lv_conv3dmask *k = new lv_conv3dmask();
  lv3->set_domain(flt); lv3->set_range(dat);
  k->set_domain(flt); k->set_range(flt);
  chain_op *dk = new chain_op(k,lv3);

  /* Create the identity operator */
  identity *iop = new identity();
  iop->set_domain(flt); iop->set_range(flt);

  /* Run the regularized inversion */
  conj_grad cg = conj_grad(niter);
  if(!rmov.empty()) {
    cg.set_resout("rmov");
  } if(!gmov.empty()) {
    cg.set_grdout("gmov");
  } if(!mmov.empty()) {
    cg.set_modout("mmov");
  }
  cg.run(dk,iop,dat,zro,flt,eps);
  lv3->set_zerolag(flt);
  sregf pefout = sregf(flt->get_axes());
  pefout.write_description("out");
  pefout.write_all("out",flt);

  /* Write out filter parameters */
  printf("j=%d,%d,%d\n",j[0],j[1],j[2]);
  if(!nlag.empty()) {
    printf("nlag=%d,%d,%d",nlag[0],nlag[1],nlag[2]);
  } else {
    printf("lag1=");
    for(int il = 0; il < (int)lag1.size(); ++il) {
      printf("%d,",lag1[il]);
    }
    printf("\n");
    printf("lag2=");
    for(int il = 0; il < (int)lag2.size(); ++il) {
      printf("%d,",lag2[il]);
    }
    printf("\n");
    printf("lag3=");
    for(int il = 0; il < (int)lag3.size(); ++il) {
      printf("%d,",lag3[il]);
    }
  }
}
