/**
 * Model 2D multicomponent shot gathers
 * given an input source, velocity model
 * and acquisition parameters
 * @author: Joseph Jennings
 * @version: 2017.04.05
 */

#include "sep_params.h"
#include "sourcescale.h"
#include "zeropad.h"
#include "laplacian10th.h"
#include "laplacian10thispc.h"
#include "afd2dfst.h"
#include "../main/mcrrec.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);
  int verb = pars.get_int("verb",0);
  int wfldwj = pars.get_int("wfldwj",1);
  /* Source axes */
  int nsx    = pars.get_int  ("nsx",  1);
  float osx  = pars.get_float("osx",0.0);
  float dsx  = pars.get_float("dsx",1.0);
  float srcz = pars.get_float("srcz",50.0);
  std::vector<float> sxs = pars.get_vfloats("sxs",10);
  std::vector<float> szs = pars.get_vfloats("szs",10);
  if(!sxs.empty() && !szs.empty()) nsx = sxs.size();
  /* Receiver axes */
  int nrx    = pars.get_int  ("nrx",  1);
  float orx  = pars.get_float("orx",0.0);
  float drx  = pars.get_float("drx",1.0);
  float recz = pars.get_float("recz",50.0);
  /* Wavefield sampling */
  float dtu = pars.get_float("dtu",0.001);
  float fmax = pars.get_float("fmax",50);
  /* Density at receivers */
  float rho = pars.get_float("rho",1000.0);

  /* Output file for wavefields */
  std::string wfname = pars.get_string("wfldsout","");

  /* Input files */
  sregf srcin = sregf("in");
  sregf velin = sregf("vel");
  sregf wfout;

  int   nz = velin.get_axis(1).n;
  float dz = velin.get_axis(1).d;
  int   nx = velin.get_axis(2).n;
  float dx = velin.get_axis(2).d;
  int   nt = srcin.get_axis(1).n;

  std::vector<axis> daxes,raxes;
  daxes.push_back(srcin.get_axis(1)); raxes.push_back(srcin.get_axis(1));
  daxes.push_back(axis(nrx,orx,drx)); raxes.push_back(axis(nrx,orx,drx));
  daxes.push_back(axis(3,0.0,1.0));   raxes.push_back(axis(3,0.0,1.0));
  daxes.push_back(axis(nsx,osx,dsx));

  std::vector<axis> paxes,caxes;
  /* Wavefield axes */
  paxes.push_back(axis(nz,0.0,dz));
  paxes.push_back(axis(nx,0.0,dx));
  paxes.push_back(srcin.get_axis(1));
  /* Multiple component axes */
  caxes = paxes;
  caxes.push_back(axis(2,0.0,1.0));

  /* Build hypercubes */
  hypercube_float *vel = new hypercube_float(velin.get_axes(),true);
  hypercube_float *src = new hypercube_float(srcin.get_axes(),true);
  hypercube_float *dat = new hypercube_float(daxes,true);

  /* Read in input files */
  srcin.read_all("in",src);
  velin.read_all("vel",vel);
  if(!wfname.empty()) {
    wfout = sregf(paxes);
    wfout.write_description(wfname);
  }

  /* Construct operators */
  sourcescale ss          = sourcescale(dtu,vel);

  /* Loop over shot */
  int iw = 0;
#pragma omp parallel for default(shared)
  for(int is = 0; is < nsx; ++is) {
    laplacian10thispc *lp   = new laplacian10thispc(paxes);
    afd2dfst fd             = afd2dfst(paxes,vel,lp,dtu,fmax);
    mcrrec mcrr             = mcrrec(paxes,axis(nrx,orx,drx),recz, rho);
    hypercube_float *prc = new hypercube_float(paxes,true);
    hypercube_float *ssc = new hypercube_float(paxes,true);
    hypercube_float *rec = new hypercube_float(raxes,true);
    float xs = osx + dsx*is;
    if(verb) fprintf(stderr,"shotnum=%d shotpos=%f\n",is,xs);
    /* Use coordinate if provided */
    if(!sxs.empty() && !szs.empty()) {
      xs = sxs[is]; srcz = szs[is];
    }
    zeropad zp = zeropad(static_cast<int>(xs),static_cast<int>(srcz));
    zp.forward(false,src,prc);
    ss.forward(false,prc,ssc);
    fd.forward(false,ssc,prc);
    /* If user specifies, write out wavefield */
    if(!wfname.empty()) {
      if(is%wfldwj == 0) {
        wfout.write_all(wfname,prc);
        wfout.write_axis(wfname,4,axis(iw+1,0,1.0,"shot"));
        ++iw;
      }
    }
    /* Multicomponent receiver restriction */
    mcrr.adjoint(false,rec,prc);
    memcpy(&dat->vals[3*is*nt*nrx],rec->vals,sizeof(float)*3*nt*nrx);
    delete prc; delete ssc; delete rec;
  }

  /* Write out data */
  sregf datout = sregf(dat->get_axes());
  datout.write_description("out");
  datout.write_all("out",dat);

  /* Clean up */
  delete vel; delete src;
}
