/**
 * Applies a smoothly-varying 2C or 3C rotation
 * operator to a multicomponent time series.
 * Assumes vector order [z,x,y]
 * @author: Joseph Jennings
 * @version: 2018.02.06
 */

#include <math.h>
#include "hypercube_float.h"
#include "sregf.h"
#include "../main/config_parser.h"
#include "../main/error_handle.h"
#include "../main/lininterp.h"
#include "../main/simpleplot.h"

# define DEG2RAD M_PI/180.0

int main(int argc, char **argv) {

  /* Get input parameters */
  ConfigParser cfg = ConfigParser(argc, argv, true);
  bool verb = cfg.get_int("cmd", "verb", false);
  /* Rotation about y, x and z */
  std::vector<float> thetay; int nthetay = 0;
  std::vector<float> thetax; int nthetax = 0; bool nothetax = false;
  std::vector<float> thetaz; int nthetaz = 0;
  try {
    thetay = cfg.get_floats("cmd", "thetay");
  } catch (const InvalidArg &ex) {
    fprintf(stderr,"Must provide at least one angle for rotation.\n");
    return 1;
  }
  try {
    thetax = cfg.get_floats("cmd","thetax");
  } catch(const InvalidArg &ex) {
    nothetax = true;
  }
  try {
      thetaz = cfg.get_floats("cmd", "thetaz");
  } catch(const InvalidArg &ex) {
    if(verb && nothetax) fprintf(stderr, "No thetax or thetaz provided, only rotating 2 components.\n");
  }
  std::string angout = cfg.get("cmd","angout","");

  nthetay = (int)thetay.size();
  nthetax = (int)thetax.size();
  nthetaz = (int)thetaz.size();

  /* Get input MC time series */
  sregf mcin = sregf("in");

  int nt = mcin.get_axis(1).n;
  int nc = mcin.get_axis(2).n;

  /* Coarse and fine theta axes */
  std::vector<axis> crsaxes;
  float dtheta = static_cast<float>(nt)/static_cast<float>(nthetay-1.0);
  crsaxes.push_back(axis(nthetay,0.0,dtheta));
  for(int icrs = 0; icrs < nthetay; ++icrs) {
    fprintf(stderr,"crs=%f thetay=%f\n",dtheta*icrs,thetay[icrs]);
  }
  std::vector<axis> finaxes;
  finaxes.push_back(axis(nt,0.0,1.0));

  /* Read in data */
  hypercube_float *mcdat   = new hypercube_float(mcin.get_axes(),true);
  hypercube_float *rotdat  = new hypercube_float(mcin.get_axes(),true);
  hypercube_float *thetaycrs = new hypercube_float(crsaxes,thetay.data());
  hypercube_float *thetayfin = new hypercube_float(finaxes,true);

  mcin.read_all("in", mcdat);

  if(nc == 2 || (nthetax == 0 && nthetaz == 0)) {
    if(verb) fprintf(stderr,"Two component data, performing two component rotation.\n");
    /* Linear interpolation of the angles */
    lininterp li = lininterp(1);
    li.forward(false, thetaycrs, thetayfin);
    /* Apply rotation operator with changing angles */
    for(int it = 0; it < nt; ++it) {
      rotdat->vals[(0)*nt + it] = cos(thetayfin->vals[it]*DEG2RAD)*mcdat->vals[(0)*nt + it] -
          sin(thetayfin->vals[it]*DEG2RAD)*mcdat->vals[(1)*nt + it];
      rotdat->vals[(1)*nt+ it] = sin(thetayfin->vals[it]*DEG2RAD)*mcdat->vals[(0)*nt + it] +
          cos(thetayfin->vals[it]*DEG2RAD)*mcdat->vals[(1)*nt + it];
    }
  } else {
  }

  if(!angout.empty()){
    sregf ang = sregf(thetayfin->get_axes());
    ang.write_description("angout");
    ang.write_all("angout", thetayfin);
  }
  /* Write out rotated data */
  sregf rotout = sregf(rotdat->get_axes());
  rotout.write_description("out");
  rotout.write_all("out", rotdat);

  /* Clean up */
  delete mcdat; delete rotdat; delete thetaycrs; delete thetayfin;
}
