/**
 * Separate simultaneous source blended data
 * using 1D signal and noise PEFs estimated
 * on the independently-modeled, proxy
 * data.
 * @author: Joseph Jennings
 * @version: 2017.12.21
 */

#include "sep_params.h"
#include "super_vector.h"
#include "combo_oper.h"
#include "../main/lv_conv1dm.h"
#include "../main/identity.h"
#include "../main/conj_grad.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc, argv);
  int niter = pars.get_int("niter");
  int vrb   = pars.get_int("verb",1);
  std::string nfile = pars.get_string("noizout","");
  std::string rmov = pars.get_string("rmov","");
  std::string smmov = pars.get_string("smmov","");
  std::string nmmov = pars.get_string("nmmov","");
  std::string sgmov = pars.get_string("sgmov","");
  std::string ngmov = pars.get_string("ngmov","");
  if(nfile.empty()) {
    seperr("Must provide a noise file for writing out inverted noise model\n");
  }

  /* Read in data and PEFs */
  sregf datin  = sregf("in");
  sregf spefin = sregf("spef");
  int js    = pars.get_auxint("j","spef");
  int nlags = pars.get_auxint("nlag","spef",0);
  std::vector<int> lag1s = pars.get_auxvints("lag1",50,"spef");
  sregf npefin = sregf("npef");
  int jn = pars.get_auxint("j","npef");
  int nlagn = pars.get_auxint("nlag","npef",0);
  std::vector<int> lag1n = pars.get_auxvints("lag1",50,"npef");

  hypercube_float *spef  = new hypercube_float(spefin.get_axes(),true);
  hypercube_float *npef  = new hypercube_float(npefin.get_axes(),true);
  hypercube_float *dat   = new hypercube_float(datin.get_axes(),true);
  hypercube_float *wdats = new hypercube_float(datin.get_axes(),true);
  hypercube_float *wdatn = new hypercube_float(datin.get_axes(),true);
  hypercube_float *sig   = new hypercube_float(datin.get_axes(),true);
  hypercube_float *noiz  = new hypercube_float(datin.get_axes(),true);
  hypercube_float *bld   = new hypercube_float(datin.get_axes(),true);
  hypercube_float *zero  = new hypercube_float(datin.get_axes(),true);
  spefin.read_all("spef",spef);
  npefin.read_all("npef",npef);
  datin.read_all("in",dat);

  int nd = dat->get_axis(1).n;

  /* Creation of signal PEF */
  lv_conv1dm *slv = nullptr;
  if(nlags != 0) {
    slv = new lv_conv1dm(nd,js,nlags,spef);
  } else {
    slv = new lv_conv1dm(nd,js,lag1s,spef);
    if(vrb) {
      fprintf(stderr,"Signal PEF:\n");
      slv->print_laginfo();
      fprintf(stderr,"\n");
      slv->print_blockinfo();
      fprintf(stderr,"\n");
    }
  }

  /* Creation of noise PEF */
  lv_conv1dm *nlv = nullptr;
  if(nlagn != 0) {
    nlv = new lv_conv1dm(nd,jn,nlagn,npef);
  } else {
    nlv = new lv_conv1dm(nd,js,lag1n,npef);
    if(vrb) {
      fprintf(stderr,"Noise PEF:\n");
      nlv->print_laginfo();
      fprintf(stderr,"\n");
      nlv->print_blockinfo();
    }
  }

  /* Creation of blending operator */
  identity *iop1 = new identity();
  identity *iop2 = new identity();
  iop1->set_domain(dat); iop1->set_range(dat);
  iop2->set_domain(dat); iop2->set_range(dat);
  row_op *gam = new row_op(iop1,iop2);

  /* Diagonal PEF operator */
  slv->set_domain(wdatn); slv->set_range(noiz);
  nlv->set_domain(wdats); nlv->set_range(sig);
  diag_op *pefs = new diag_op(nlv,slv);

  /* Create weighted data */
  slv->forward(false,dat,wdats);
  nlv->forward(false,dat,wdatn);

  /* Create model and data vectors */
  super_vector *sn = new super_vector(sig,noiz);

  /* Build total operator */
  super_vector *wdat = new super_vector(wdatn,wdats);
  super_vector *datinv = new super_vector(dat,wdat);
  col_op *A = new col_op(gam,pefs);

  /* Run inversion */
  conj_grad cg = conj_grad(niter);
  if(!rmov.empty()) {
    cg.set_resout("rmov");
  } if(!sgmov.empty()) {
    cg.set_grd1out("sgmov");
  } if(!ngmov.empty()) {
    cg.set_grd2out("ngmov");
  } if(!smmov.empty()) {
    cg.set_mod1out("smmov");
  } if(!nmmov.empty()) {
    cg.set_mod2out("nmmov");
  }
  cg.run(A,datinv,sn);

  /* Write out signal */
  sregf sigout = sregf(sig->get_axes());
  sigout.write_description("out");
  sigout.write_all("out",(hypercube_float*)sn->vecs[0]);

  /* Write out noise */
  sregf nozout = sregf(noiz->get_axes());
  nozout.write_description("noizout");
  nozout.write_all("noizout",(hypercube_float*)sn->vecs[1]);

  /* Clean up */
  delete bld; delete wdats; delete wdatn; delete zero;
  delete spef; delete npef; delete sig; delete noiz;
}
