/**
 * Tests the scalar time-varying
 * PEF code that Jon has in his
 * vector deconvolution paper
 * @author: Joseph Jennings
 * @version: 2017.05.25
 */

#include <math.h>
#include "sep_params.h"
#include "../main/simpleplot.h"

int main(int argc, char **argv) {

  sep_params pars = sep_params(argc,argv);
  int verb     = pars.get_int("verb",0);
  int gap      = pars.get_int("gap",1);
  int na       = pars.get_int("na",10);
  int lambda   = pars.get_int("lambda",200);
  float eps    = (float)1/lambda;
  std::string flttag = pars.get_string("fflt");

  sregf in = sregf("in");
  int nt = in.get_axis(1).n;

  std::vector<axis> faxes;
  faxes.push_back(axis(na,0.0,1.0));

  hypercube_float *y   = new hypercube_float(in.get_axes(),true);
  hypercube_float *A   = new hypercube_float(faxes,true);
  hypercube_float *e   = new hypercube_float(in.get_axes(),true);
  A->zero(); e->zero();

  /* Constrain first coefficient to be one (PEF) */
  A->vals[0] = 1.0;

  /* Read in y */
  in.read_all("in",y);

  /* Initialize scaling terms (sigmay and simgae) */
  float sumsq = 0.0;
  //FIXME: First problem is how to initialize these guys
  for(int it = 0; it < nt; ++it) {
    sumsq += y->vals[it]*y->vals[it];
  }
  float sigy = sqrt(sumsq/(float)lambda);
  float sige = sigy/2;

  if(verb) {
    fprintf(stderr,"sigy=%f sige=%f\n",sigy,sige);
    fprintf(stderr,"lambda=%d epsilon=%f\n",lambda,eps);
  }
  /* Loop over all predictable samples */
  for(int it = na; it < nt; ++it) {
    e->vals[it] = 0.0;
    /* Perform forward convolution */
    for(int itau = 0; itau < na; ++itau) {
      e->vals[it] += A->vals[itau]*y->vals[it-itau];
    }
    /* Compute running variance */
    sige = (1.0-eps) * sige + eps * sqrt(e->vals[it]*e->vals[it]);
    sigy = (1.0-eps) * sigy + eps * sqrt(y->vals[it]*y->vals[it]);
    /* Perform adjoint convolution (correlation) */
    for(int itau = gap; itau < na; ++itau) {
      A->vals[itau] -= eps/(sige*sigy) * e->vals[it] * y->vals[it-itau];
    }
  }

  //simpleplot::plot1d2(y,e,true);

  /* Output error */
  sregf eout = sregf(e->get_axes());
  eout.write_description("out");
  eout.write_all("out",e);

  /* Output final filter */
  if(!flttag.empty()) {
    sregf aout = sregf(A->get_axes());
    aout.write_description("fflt");
    aout.write_all("fflt",A);
  }

  /* Clean up */
  delete y; delete A; delete e;
}
