/**
 * Tests the dot product and other
 * simple properties of the multichannel
 * filter
 * @author: Joseph Jennings
 * @version: 2017.05.03
 */

#include <limits.h>
#include "gtest/gtest.h"
#include "axis.h"
#include "hypercube_float.h"
#include "../main/mctcai.h"
#include "../main/tcai.h"
#include "../main/identity.h"
#include "../main/simpleplot.h"

class mctcaitest: public ::testing::Test {
  public:
  protected:
    virtual void SetUp(){
    }

    virtual void TearDown() {
    }
};

TEST_F(mctcaitest,sumtest) {

  int n = 100;

  std::vector<axis> axes;
  axes.push_back(axis(n,0.0,1.0));

  hypercube_float *one = new hypercube_float(axes,true);
  one->set_val(1.0);
  hypercube_float *two = new hypercube_float(axes,true);
  two->set_val(2.0);
  hypercube_float *th1 = new hypercube_float(axes,true);
  hypercube_float *th2 = new hypercube_float(axes,true);

  std::vector<my_operator*> iops;
  identity *iop11 = new identity();
  identity *iop12 = new identity();
  identity *iop21 = new identity();
  identity *iop22 = new identity();
  iop11->set_range(one); iop11->set_domain(one);
  iop12->set_range(one); iop12->set_domain(one);
  iop21->set_range(one); iop21->set_domain(one);
  iop22->set_range(one); iop22->set_domain(one);

  iops.push_back(iop11);
  iops.push_back(iop12);
  iops.push_back(iop21);
  iops.push_back(iop22);

  mctcai mciop = mctcai(iops);

  super_vector *mod = new super_vector(one,two);
  super_vector *dat = new super_vector(th1,th2);

  mciop.forward(false,mod,dat);
  ASSERT_EQ(dat->vecs[0]->sum()/n,3.000);
  ASSERT_EQ(dat->vecs[1]->sum()/n,3.000);

}

TEST_F(mctcaitest,idtest) {

  int n = 100;

  std::vector<axis> axes;
  axes.push_back(axis(n,0.0,1.0));

  hypercube_float *one = new hypercube_float(axes,true);
  one->set_val(1.0);
  hypercube_float *two = new hypercube_float(axes,true);
  two->set_val(2.0);
  hypercube_float *th1 = new hypercube_float(axes,true);
  hypercube_float *th2 = new hypercube_float(axes,true);

  std::vector<my_operator*> iops;
  identity *iop11 = new identity();
  identity *iop12 = new identity();
  identity *iop21 = new identity();
  identity *iop22 = new identity();
  iop11->set_range(one); iop11->set_domain(one);
  iop12->set_range(one); iop12->set_domain(one);
  iop21->set_range(one); iop21->set_domain(one);
  iop22->set_range(one); iop22->set_domain(one);

  iops.push_back(iop11);
  iops.push_back(iop12);
  iops.push_back(iop21);
  iops.push_back(iop22);
  iops[1]->set_scale(0.0);
  iops[2]->set_scale(0.0);

  mctcai mciop = mctcai(iops);

  super_vector *mod = new super_vector(one,two);
  super_vector *dat = new super_vector(th1,th2);

  mciop.forward(false,mod,dat);
  ASSERT_EQ(dat->vecs[0]->sum()/n,1.000);
  ASSERT_EQ(dat->vecs[1]->sum()/n,2.000);

}

TEST_F(mctcaitest,dptest) {

  int nm = 100; int nf = 10;
  int nd = nm + nf - 1;

  /* Build axes */
  axis af = axis(nf,0.0,1.0);
  axis am = axis(nm,0.0,1.0);
  axis ad = axis(nd,0.0,1.0);
  std::vector<axis> maxes, daxes, faxes;
  faxes.push_back(af);
  maxes.push_back(am); daxes.push_back(ad);
  /* Model */
  hypercube_float *m1  = new hypercube_float(maxes,true);
  hypercube_float *m2  = new hypercube_float(maxes,true);
  hypercube_float *m1s = new hypercube_float(maxes,true);
  hypercube_float *m2s = new hypercube_float(maxes,true);
  m1->random(); m2->random();
  /* Data */
  hypercube_float *d1  = new hypercube_float(daxes,true);
  hypercube_float *d2  = new hypercube_float(daxes,true);
  hypercube_float *d1s = new hypercube_float(daxes,true);
  hypercube_float *d2s = new hypercube_float(daxes,true);
  d1->random(); d2->random();
  /* Filters */
  hypercube_float *f11 = new hypercube_float(faxes,true);
  hypercube_float *f12 = new hypercube_float(faxes,true);
  hypercube_float *f21 = new hypercube_float(faxes,true);
  hypercube_float *f22 = new hypercube_float(faxes,true);
  f11->random(); f12->random();
  f21->random(); f22->random();

  /* Create vector of operators */
  tcai *tc11 = new tcai(f11);
  tcai *tc12 = new tcai(f12);
  tcai *tc21 = new tcai(f21);
  tcai *tc22 = new tcai(f22);
  tc11->set_domain(m1); tc11->set_range(d1);
  tc12->set_domain(m2); tc12->set_range(d1);
  tc21->set_domain(m1); tc21->set_range(d2);
  tc22->set_domain(m2); tc22->set_range(d2);
  std::vector<my_operator*> fops;
  fops.push_back(tc11); fops.push_back(tc12);
  fops.push_back(tc21); fops.push_back(tc22);

  /* Create multichannel filter */
  mctcai mcf = mctcai(fops);

  /* Create model and data vectors */
  super_vector *m  = new super_vector(m1 ,m2 );
  super_vector *ms = new super_vector(m1s,m2s);

  super_vector *d  = new super_vector(d1, d2);
  super_vector *ds = new super_vector(d1s, d2s);

  /* Add == false */
  mcf.forward(false,m,ds);
  double dotd = d->dot(ds);
  mcf.adjoint(false,ms,d);
  double dotm = m->dot(ms);
  fprintf(stderr,"%f %f\n",dotm,dotd);
  ASSERT_NEAR(dotd,dotm,0.000001);

  /* Add == true */
  mcf.forward(true,m,ds);
  double adotd = d->dot(ds);
  mcf.adjoint(true,ms,d);
  double adotm = m->dot(ms);
  fprintf(stderr,"%f %f\n",adotm,adotd);
  ASSERT_NEAR(adotd,adotm,0.000001);

}
